#
# Copyright (C) 2020 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import typing

import attr
import numpy as np

import fwdpy11._fwdpy11

from .class_decorators import (
    attr_add_asblack,
    attr_class_pickle_with_super,
    attr_class_to_from_dict,
    attr_class_to_from_dict_no_recurse,
    region_custom_repr,
)


@region_custom_repr
@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict
@attr.s(auto_attribs=True, frozen=True, repr=False, repr_ns="fwdpy11")
class Region(fwdpy11._fwdpy11._ll_Region):
    """
    A genomic region, defined by half-open interval [beg, end)

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param beg: the beginning of the region
    :type beg: float
    :param end: the end of the region
    :type end: float
    :param weight: the weight to assign
    :type weight: float
    :param coupled: if True, the weight is converted to (end-beg)*weight
    :type coupled: bool
    :param label: Not relevant to recombining regions.
                  Otherwise, this value will be used to fill
                  :attr:`fwdpy11.Mutation.label`.
    :type label: :attr:`numpy.uint16`

    When coupled is True, the "weight" may be interpreted
    as a "per base pair" (or per unit, generally speaking) term.

    .. versionchanged:: 0.3.0

        Refactored from a pure Python class to a C++/pybind11 class

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    beg: float
    end: float
    weight: float
    coupled: bool = attr.ib(default=True)
    label: int = attr.ib(default=0)

    def __attrs_post_init__(self):
        super(Region, self).__init__(
            self.beg, self.end, self.weight, self.coupled, self.label
        )


@region_custom_repr
@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict_no_recurse
@attr.s(auto_attribs=True, frozen=True, repr=False, repr_ns="fwdpy11")
class ConstantS(fwdpy11._fwdpy11._ll_ConstantS):
    """
    Mutations with fixed effect sizes

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param beg: the beginning of the region
    :type beg: float
    :param end: the end of the region
    :type end: float
    :param weight: the weight to assign
    :type weight: float
    :param s: the selection coefficient
    :type s: float
    :param h: the dominance
    :type h: typing.Union[float, fwdpy11.MutationDominance]
    :param coupled: if True, the weight is converted to (end-beg)*weight
    :type coupled: bool
    :param label: Fill :attr:`fwdpy11.Mutation.label` with this value.
    :type label: :attr:`numpy.uint16`
    :param scaling: The scaling of the DFE
    :type scaling: float

    When coupled is True, the "weight" may be interpreted
    as a "per base pair" (or per unit, generally speaking) term.

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    beg: float
    end: float
    weight: float
    s: float
    h: float = attr.ib(default=1.0)
    coupled: bool = attr.ib(default=True)
    label: int = attr.ib(default=0)
    scaling: float = attr.ib(default=1.0)

    def __attrs_post_init__(self):
        super(ConstantS, self).__init__(
            self.beg,
            self.end,
            self.weight,
            self.s,
            self.h,
            self.coupled,
            self.label,
            self.scaling,
        )


@region_custom_repr
@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict_no_recurse
@attr.s(auto_attribs=True, frozen=True, repr=False, repr_ns="fwdpy11")
class ExpS(fwdpy11._fwdpy11._ll_ExpS):
    """
    Exponential distribution of effect sizes

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param beg: the beginning of the region
    :type beg: float
    :param end: the end of the region
    :type end: float
    :param weight: the weight to assign
    :type weight: float
    :param mean: the mean selection coefficient
    :type s: float
    :param h: the dominance
    :type h: typing.Union[float, fwdpy11.MutationDominance]
    :param coupled: if True, the weight is converted to (end-beg)*weight
    :type coupled: bool
    :param label: Fill :attr:`fwdpy11.Mutation.label` with this value.
    :type label: :attr:`numpy.uint16`
    :param scaling: The scaling of the DFE
    :type scaling: float

    When coupled is True, the "weight" may be interpreted
    as a "per base pair" (or per unit, generally speaking) term.

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    import numpy as np

    beg: float
    end: float
    weight: float
    mean: float
    h: float = attr.ib(default=1.0)
    coupled: bool = attr.ib(default=True)
    label: int = attr.ib(default=0)
    scaling: float = attr.ib(default=1.0)

    def __attrs_post_init__(self):
        super(ExpS, self).__init__(
            self.beg,
            self.end,
            self.weight,
            self.mean,
            self.h,
            self.coupled,
            self.label,
            self.scaling,
        )


@region_custom_repr
@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict_no_recurse
@attr.s(auto_attribs=True, frozen=True, repr=False, repr_ns="fwdpy11")
class GammaS(fwdpy11._fwdpy11._ll_GammaS):
    """
    Gamma distribution of effect sizes

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param beg: the beginning of the region
    :type beg: float
    :param end: the end of the region
    :type end: float
    :param weight: the weight to assign
    :type weight: float
    :param mean: the mean selection coefficient
    :type mean: float
    :param shape_parameter: the shape parameter of the distribution
    :type shape_parameter: float
    :param h: the dominance
    :type h: typing.Union[float, fwdpy11.MutationDominance]
    :param coupled: if True, the weight is converted to (end-beg)*weight
    :type coupled: bool
    :param label: Fill :attr:`fwdpy11.Mutation.label` with this value.
    :type label: :attr:`numpy.uint16`
    :param scaling: The scaling of the DFE
    :type scaling: float

    When coupled is True, the "weight" may be
    interpreted as a "per base pair" (or per unit, generally speaking) term.

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    beg: float
    end: float
    weight: float
    mean: float
    shape_parameter: float
    h: float = 1.0
    coupled: bool = True
    label: int = 0
    scaling: float = 1.0

    def __attrs_post_init__(self):
        super(GammaS, self).__init__(
            self.beg,
            self.end,
            self.weight,
            self.mean,
            self.shape_parameter,
            self.h,
            self.coupled,
            self.label,
            self.scaling,
        )


@region_custom_repr
@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict_no_recurse
@attr.s(auto_attribs=True, frozen=True, repr=False, repr_ns="fwdpy11")
class GaussianS(fwdpy11._fwdpy11._ll_GaussianS):
    """
    Gaussian distribution of effect sizes

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param beg: the beginning of the region
    :type beg: float
    :param end: the end of the region
    :type end: float
    :param weight: the weight to assign
    :type weight: float
    :param sd: standard deviation of effect sizes
    :type sd: float
    :param h: the dominance
    :type h: typing.Union[float, fwdpy11.MutationDominance]
    :param coupled: if True, the weight is converted to (end-beg)*weight
    :type coupled: bool
    :param label: Fill :attr:`fwdpy11.Mutation.label` with this value.
    :type label: :attr:`numpy.uint16`
    :param scaling: The scaling of the DFE
    :type scaling: float

    When coupled is True, the "weight" may be
    interpreted as a "per base pair" (or per unit, generally speaking) term.

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    beg: float
    end: float
    weight: float
    sd: float
    h: float = 1.0
    coupled: bool = True
    label: int = 0
    scaling: float = 1.0

    def __attrs_post_init__(self):
        super(GaussianS, self).__init__(
            self.beg,
            self.end,
            self.weight,
            self.sd,
            self.h,
            self.coupled,
            self.label,
            self.scaling,
        )


@region_custom_repr
@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict_no_recurse
@attr.s(auto_attribs=True, frozen=True, repr=False, repr_ns="fwdpy11")
class LogNormalS(fwdpy11._fwdpy11._ll_LogNormalS):
    """
    Log-normal distribution of effect sizes.

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param beg: the beginning of the region
    :type beg: float
    :param end: the end of the region
    :type end: float
    :param weight: the weight to assign
    :type weight: float
    :param zeta: the zeta parameter
    :type zeta: float
    :param sigma: the sigma parameter
    :type sigma: float
    :param h: the dominance
    :type h: typing.Union[float, fwdpy11.MutationDominance]
    :param coupled: if True, the weight is converted to(end-beg)*weight
    :type coupled: bool
    :param label: Fill :attr:`fwdpy11.Mutation.label` with this value.
    :type label: :attr:`numpy.uint16`
    :param scaling: The scaling of the DFE
    :type scaling: float

    When coupled is True, the "weight" may be
    interpreted as a "per base pair" (or per unit, generally speaking) term.

    .. versionadded:: 0.7.0

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    beg: float
    end: float
    weight: float
    zeta: typing.Optional[float]
    sigma: typing.Optional[float]
    h: float = 1.0
    coupled: bool = True
    label: int = 0
    scaling: float = 1.0

    def __attrs_post_init__(self):
        super(LogNormalS, self).__init__(
            self.beg,
            self.end,
            self.weight,
            self.zeta,
            self.sigma,
            self.h,
            self.coupled,
            self.label,
            self.scaling,
        )

    @classmethod
    def mv(
        cls,
        beg: float,
        end: float,
        weight: float,
        h: float = 1.0,
        coupled: bool = True,
        label: int = 0,
        scaling: float = 1.0,
    ):
        """
        Factory method to create an instance compatible with
        :class:`fwdpy11.mvDES`. See :ref:`mvdes` for
        details.

        :param beg: the beginning of the region
        :type beg: float
        :param end: the end of the region
        :type end: float
        :param weight: the weight to assign
        :type weight: float
        :param h: the dominance
        :type h: typing.Union[float, fwdpy11.MutationDominance]
        :param coupled: if True, the weight is converted to(end-beg)*weight
        :type coupled: bool
        :param label: Fill :attr:`fwdpy11.Mutation.label` with this value.
        :type label: :attr:`numpy.uint16`
        :param scaling: The scaling of the DFE
        :type scaling: float
        """

        return cls(
            beg=beg,
            end=end,
            weight=weight,
            zeta=None,
            sigma=None,
            h=h,
            coupled=coupled,
            label=label,
            scaling=scaling,
        )


@region_custom_repr
@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict_no_recurse
@attr.s(auto_attribs=True, frozen=True, repr=False, repr_ns="fwdpy11")
class UniformS(fwdpy11._fwdpy11._ll_UniformS):
    """
    Uniform distrubution of effect sizes

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param beg: the beginning of the region
    :type beg: float
    :param end: the end of the region
    :type end: float
    :param weight: the weight to assign
    :type weight: float
    :param lo: lower bound on s
    :type lo: float
    :param hi: upper bound on s
    :type hi: float
    :param h: the dominance
    :type h: typing.Union[float, fwdpy11.MutationDominance]
    :param coupled: if True, the weight is converted to (end-beg)*weight
    :type coupled: bool
    :param label: Fill :attr:`fwdpy11.Mutation.label` with this value.
    :type label: :attr:`numpy.uint16`
    :param scaling: The scaling of the DFE
    :type scaling: float

    When coupled is True, the "weight" may be
    interpreted as a "per base pair" (or per unit, generally speaking) term.

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    beg: float
    end: float
    weight: float
    lo: float
    hi: float
    h: float = 1.0
    coupled: bool = True
    label: int = 0
    scaling: float = 1.0

    def __attrs_post_init__(self):
        super(UniformS, self).__init__(
            self.beg,
            self.end,
            self.weight,
            self.lo,
            self.hi,
            self.h,
            self.coupled,
            self.label,
            self.scaling,
        )


@region_custom_repr
@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict_no_recurse
@attr.s(eq=False, auto_attribs=True, frozen=True, repr=False, repr_ns="fwdpy11")
class MultivariateGaussianEffects(fwdpy11._fwdpy11._ll_MultivariateGaussianEffects):
    """
    Pleiotropic effects via a multivariate Gaussian distribution.

    This class can be used to generate mutations with both vectors
    of effect sizes as well as a separate fixed effect.

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param beg: Beginning of the region
    :type beg: float
    :param end: End of the region
    :type end: float
    :param weight: Weight on the region
    :type weight: float
    :param cov_matrix: Variance-covariance matrix
    :type cov_matrix: :class:`numpy.ndarray`
    :param fixed_effect: Fixed effect size. Defaults to 0.0.
    :type fixed_effect: float
    :param h: Dominance. Defaults to 1.0
    :type h: typing.Union[float, fwdpy11.MutationDominance]
    :param coupled: Specify if weight is function of end-beg or not. Defaults to True
    :type coupled: bool
    :param label: Fill :attr:`fwdpy11.Mutation.label` with this value.
    :type label: :attr:`numpy.uint16`

    The input matrix must be square and semi-positive definite.   If either
    of these conditions are not met, ValueError will be raised. ValueError
    will also be raised if the input matrix contains any non-finite values.

    .. note::

        The dominance parameter (`h`) applies to both the fixed effect and those
        drawn from a multivariate normal.

    .. versionadded:: 0.3.0

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """

    beg: float
    end: float
    weight: float
    cov_matrix: np.ndarray
    fixed_effect: float = 0.0
    h: float = 1.0
    coupled: bool = True
    label: int = 0

    def __attrs_post_init__(self):
        super(MultivariateGaussianEffects, self).__init__(
            self.beg,
            self.end,
            self.weight,
            self.cov_matrix,
            self.fixed_effect,
            self.h,
            self.coupled,
            self.label,
        )

    def __eq__(self, other):
        if self.beg != other.beg:
            return False

        if self.beg != other.beg:
            return False
        if self.end != other.end:
            return False
        if self.weight != other.weight:
            return False
        if self.fixed_effect != other.fixed_effect:
            return False
        if self.h != other.h:
            return False
        if self.coupled != other.coupled:
            return False
        if self.label != other.label:
            return False
        return np.array_equal(self.cov_matrix, other.cov_matrix)


@region_custom_repr
@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict_no_recurse
@attr.s(eq=False, auto_attribs=True, frozen=True, repr=False, repr_ns="fwdpy11")
class mvDES(fwdpy11._fwdpy11._ll_mvDES):
    """
    General multivariate distribution of effect sizes.

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param des: Distributions of effect sizes
    :type des: list or mvDES
    :param means: means marginal gaussian Distributions
    :type means: :class:`numpy.ndarray`
    :param matrix: Variance/covariance matrix
    :type matrix: :class:`numpy.ndarray`

    .. versionadded:: 0.7.0

    .. versionchanged:: 0.8.0

        Refactored to use attrs and inherit from
        low-level C++ class
    """
    des: typing.Union[typing.List, fwdpy11._fwdpy11.Sregion]
    means: np.ndarray
    matrix: typing.Optional[np.ndarray] = None

    def __attrs_post_init__(self):
        super(mvDES, self).__init__(self.des, self.means, self.matrix)


@region_custom_repr
@attr_add_asblack
@attr_class_pickle_with_super
@attr_class_to_from_dict_no_recurse
@attr.s(eq=False, auto_attribs=True, frozen=True, repr=False, repr_ns="fwdpy11")
class DiscreteDESD(fwdpy11._fwdpy11._ll_DiscreteDESD):
    """
    Discretized distribution of effect sizes and dominance.

    This class allows you to specify a discrete joint
    distrubtion of effect size and dominance.

    The distribution is specified by a list of tuples.
    Each tuple contains (effect size, dominance, weight).
    The weights must all be >= 0 and all values must fe finite.

    This class has the following attributes, whose names
    are also `kwargs` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param beg: Beginning of the region
    :type beg: float
    :param end: End of the region
    :type end: float
    :param weight: Weight on the region
    :type weight: float
    :param joint_dist: The joint distribution + weights
    :type joint_dist: list
    :param coupled: Specify if weight is function of end-beg or not. Defaults to True
    :type coupled: bool
    :param label: Fill :attr:`fwdpy11.Mutation.label` with this value.
    :type label: :attr:`numpy.uint16`
    :param scaling: The scaling of the DFE
    :type scaling: float

    .. versionadded:: 0.10.0
    """

    beg: float
    end: float
    weight: float
    joint_dist: typing.List[typing.Tuple[float, float, float]]
    coupled: bool = True
    label: int = 0
    scaling: float = 1.0

    def __attrs_post_init__(self):
        super(DiscreteDESD, self).__init__(
            self.beg,
            self.end,
            self.weight,
            self.joint_dist,
            self.coupled,
            self.label,
            self.scaling,
        )



import copy
import typing
import unittest
from dataclasses import dataclass

import demes
import fwdpy11
import numpy as np
import pytest


def run_model_round_trip(cls):
    def _test_evolvets_roundtrip(self):
        _ = evolve_demes_model(self.demog)

    cls.test_evolvets_roundtrip = _test_evolvets_roundtrip
    return cls


def resolve_and_run(b: demes.Builder, burnin: int = 1):
    g = b.resolve()
    demog = fwdpy11.discrete_demography.from_demes(g, burnin)
    evolve_demes_model(demog)


@run_model_round_trip
class TestNoEvents(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(
            description="test demography", time_units="generations")
        self.b.add_deme(name="deme", epochs=[
                        dict(start_size=1000, end_time=0)])

        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 1)


class TestBadBurnin(unittest.TestCase):
    def test_burnin_inputs(self):
        self.b = demes.Builder(
            description="test demography", time_units="generations")
        self.b.add_deme(name="deme", epochs=[
                        dict(start_size=1000, end_time=0)])

        self.g = self.b.resolve()
        with pytest.raises(ValueError):
            self.demog = fwdpy11.discrete_demography.from_demes(self.g, 10.0)
        with pytest.raises(ValueError):
            self.demog = fwdpy11.discrete_demography.from_demes(self.g, -1)


@run_model_round_trip
class TestLoadGraph(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.g = demes.load("tests/test_demog.yaml")
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 1)


@run_model_round_trip
class TestLoadYAML(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.demog = fwdpy11.discrete_demography.from_demes(
            "tests/test_demog.yaml", 1)


@run_model_round_trip
class TestTwoEpoch(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(
            description="test demography", time_units="generations")
        self.b.add_deme(
            name="deme",
            epochs=[
                dict(start_size=1000, end_time=100),
                dict(start_size=2000, end_time=0),
            ],
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 1)


@run_model_round_trip
class TestNonGenerationUnits(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(
            description="test demography", time_units="years", generation_time=25
        )
        self.b.add_deme(
            name="Pop",
            epochs=[
                dict(start_size=1000, end_time=250),
                dict(start_size=100, end_size=200, end_time=0),
            ],
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 1)


@run_model_round_trip
class TestSelfingShift(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(
            description="test demography", time_units="generations")
        self.b.add_deme(
            name="Selfer",
            epochs=[
                dict(start_size=1000, end_time=1000),
                dict(start_size=1000, end_time=0, selfing_rate=0.2),
            ],
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 1)


@run_model_round_trip
class TestSelfing(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(
            description="test demography", time_units="generations")
        self.b.add_deme(
            name="Selfer",
            epochs=[dict(start_size=1000, end_time=0)],
            defaults=dict(epoch=dict(selfing_rate=0.5)),
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 1)


@run_model_round_trip
class TestSplit(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(
            description="test demography", time_units="generations")
        self.b.add_deme(name="Ancestor", epochs=[
                        dict(start_size=1000, end_time=200)])
        self.b.add_deme(
            "Deme1", epochs=[dict(start_size=100, end_time=0)], ancestors=["Ancestor"]
        )
        self.b.add_deme(
            "Deme2", epochs=[dict(start_size=100, end_time=0)], ancestors=["Ancestor"]
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 1)


@run_model_round_trip
class TestSplitMigration(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(
            description="test demography", time_units="generations")
        self.b.add_deme(name="Ancestor", epochs=[
                        dict(start_size=1000, end_time=200)])
        self.b.add_deme(
            "Deme1", epochs=[dict(start_size=100, end_time=0)], ancestors=["Ancestor"]
        )
        self.b.add_deme(
            "Deme2", epochs=[dict(start_size=100, end_time=0)], ancestors=["Ancestor"]
        )
        self.b.add_migration(source="Deme1", dest="Deme2", rate=0.01)
        self.b.add_migration(source="Deme2", dest="Deme1", rate=0.02)
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 1)


@run_model_round_trip
class TestSplitSymmetricMigration(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(
            description="test demography", time_units="generations")
        self.b.add_deme(name="Ancestor", epochs=[
                        dict(start_size=1000, end_time=200)])
        self.b.add_deme(
            "Deme1", epochs=[dict(start_size=100, end_time=0)], ancestors=["Ancestor"]
        )
        self.b.add_deme(
            "Deme2", epochs=[dict(start_size=100, end_time=0)], ancestors=["Ancestor"]
        )
        self.b.add_migration(demes=["Deme1", "Deme2"], rate=0.01)
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 1)


@run_model_round_trip
class TestSplitAsymmetricMigration(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(
            description="test demography", time_units="generations")
        self.b.add_deme(name="Ancestor", epochs=[
                        dict(start_size=1000, end_time=200)])
        self.b.add_deme(
            "Deme1", epochs=[dict(start_size=100, end_time=0)], ancestors=["Ancestor"]
        )
        self.b.add_deme(
            "Deme2", epochs=[dict(start_size=100, end_time=0)], ancestors=["Ancestor"]
        )
        self.b.add_migration(source="Deme1", dest="Deme2", rate=0.01)
        self.b.add_migration(source="Deme2", dest="Deme1", rate=1e-3)
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 1)


@run_model_round_trip
class TestSplitThreeWay(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(
            description="test demography", time_units="generations")
        self.b.add_deme(name="Ancestor", epochs=[
                        dict(start_size=1000, end_time=200)])
        self.b.add_deme(
            "Deme1", epochs=[dict(start_size=100, end_time=0)], ancestors=["Ancestor"]
        )
        self.b.add_deme(
            "Deme2", epochs=[dict(start_size=100, end_time=0)], ancestors=["Ancestor"]
        )
        self.b.add_deme(
            "Deme3", epochs=[dict(start_size=200, end_time=0)], ancestors=["Ancestor"]
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 1)


@run_model_round_trip
class TestSplitThreeWayMigration(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(
            description="test demography", time_units="generations")
        self.b.add_deme(name="Ancestor", epochs=[
                        dict(start_size=1000, end_time=200)])
        self.b.add_deme(
            "Deme1", epochs=[dict(start_size=100, end_time=0)], ancestors=["Ancestor"]
        )
        self.b.add_deme(
            "Deme2", epochs=[dict(start_size=100, end_time=0)], ancestors=["Ancestor"]
        )
        self.b.add_deme(
            "Deme3", epochs=[dict(start_size=200, end_time=0)], ancestors=["Ancestor"]
        )
        self.b.add_migration(demes=["Deme1", "Deme2", "Deme3"], rate=0.1)
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 1)


@run_model_round_trip
class TestBranch(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="test branch",
                               time_units="generations")
        self.b.add_deme(name="Ancestor", epochs=[
                        dict(start_size=1000, end_time=0)])
        self.b.add_deme(
            "Deme1",
            epochs=[dict(start_size=100, end_time=0)],
            ancestors=["Ancestor"],
            start_time=100,
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 1)


@run_model_round_trip
class TestBranchMigration(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="test branch",
                               time_units="generations")
        self.b.add_deme(name="Ancestor", epochs=[
                        dict(start_size=1000, end_time=0)])
        self.b.add_deme(
            "Deme1",
            epochs=[dict(start_size=100, end_time=0)],
            ancestors=["Ancestor"],
            start_time=100,
        )
        self.b.add_migration(source="Ancestor", dest="Deme1", rate=0.01)
        self.b.add_migration(source="Deme1", dest="Ancestor", rate=0.005)
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 1)


@pytest.mark.parametrize("second_ancestor", ["Ancestor", "Deme1"])
@pytest.mark.parametrize("first_start_time", [100])
@pytest.mark.parametrize("second_start_time", [99])
@pytest.mark.parametrize(
    "migrations",
    [
        [],
        [
            {"source": "Ancestor", "dest": "Deme1", "rate": 0.01},
        ],
        [
            {"source": "Ancestor", "dest": "Deme2", "rate": 0.01},
        ],
        [
            {"source": "Ancestor", "dest": "Deme1", "rate": 0.01},
            {"source": "Deme1", "dest": "Ancestor", "rate": 0.005},
        ],
        [
            {"source": "Deme2", "dest": "Deme1", "rate": 0.01},
            {"source": "Deme1", "dest": "Deme2", "rate": 0.005},
        ],
        [
            {"source": "Ancestor", "dest": "Deme2", "rate": 0.01},
            {"source": "Deme2", "dest": "Ancestor", "rate": 0.005},
        ],
        [{"demes": ["Ancestor", "Deme1", "Deme2"], "rate": 1e-5}],
    ],
)
def test_three_deme_sequential_branches(
    second_ancestor, first_start_time, second_start_time, migrations
):
    b = demes.Builder(description="test branch", time_units="generations")
    b.add_deme(name="Ancestor", epochs=[dict(start_size=500, end_time=0)])
    b.add_deme(
        "Deme1",
        epochs=[dict(start_size=100, end_time=0)],
        ancestors=["Ancestor"],
        start_time=first_start_time,
    )
    b.add_deme(
        "Deme2",
        epochs=[dict(start_size=100, end_time=0)],
        ancestors=[second_ancestor],
        start_time=second_start_time,
    )
    for m in migrations:
        b.add_migration(**m)
    resolve_and_run(b)


@run_model_round_trip
class TestMultipleBranches(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="test branch",
                               time_units="generations")
        self.b.add_deme(name="Ancestor", epochs=[
                        dict(start_size=1000, end_time=0)])
        self.b.add_deme(
            "Deme1",
            epochs=[dict(start_size=100, end_time=0)],
            ancestors=["Ancestor"],
            start_time=100,
        )
        self.b.add_deme(
            "Deme2",
            epochs=[dict(start_size=200, end_time=0)],
            ancestors=["Ancestor"],
            start_time=50,
        )
        self.b.add_deme(
            "Deme3",
            epochs=[dict(start_size=300, end_time=0)],
            ancestors=["Deme1"],
            start_time=20,
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 1)


@run_model_round_trip
class TestMultipleBranchesWithMigration(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="test branch",
                               time_units="generations")
        self.b.add_deme(name="Ancestor", epochs=[
                        dict(start_size=1000, end_time=0)])
        self.b.add_deme(
            "Deme1",
            epochs=[dict(start_size=100, end_time=0)],
            ancestors=["Ancestor"],
            start_time=100,
        )
        self.b.add_deme(
            "Deme2",
            epochs=[dict(start_size=200, end_time=0)],
            ancestors=["Ancestor"],
            start_time=50,
        )
        self.b.add_deme(
            "Deme3",
            epochs=[dict(start_size=300, end_time=0)],
            ancestors=["Deme1"],
            start_time=20,
        )
        self.b.add_migration(source="Deme1", dest="Deme2", rate=1e-6)
        self.b.add_migration(source="Deme2", dest="Deme1", rate=2e-6)
        self.b.add_migration(demes=["Deme1", "Deme3"], rate=0.01)
        self.b.add_migration(source="Deme3", dest="Deme2", rate=1e-6)
        self.b.add_migration(source="Deme2", dest="Deme3", rate=2e-6)
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 1)


@run_model_round_trip
class TestSplitsBranches(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="test", time_units="generations")
        self.b.add_deme(name="A", epochs=[dict(start_size=1000, end_time=100)])
        self.b.add_deme(
            name="B",
            epochs=[dict(start_size=1000, end_time=0)],
            ancestors=["A"],
            start_time=200,
        )
        self.b.add_deme(
            name="C", epochs=[dict(start_size=1000, end_time=0)], ancestors=["A"]
        )
        self.b.add_deme(
            name="D", epochs=[dict(start_size=1000, end_time=0)], ancestors=["A"]
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 1)


@run_model_round_trip
class TestIslandModel(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="island", time_units="generations")
        self.b.add_deme(name="Island1", epochs=[
                        dict(start_size=100, end_time=0)])
        self.b.add_deme(name="Island2", epochs=[
                        dict(start_size=200, end_time=0)])
        self.b.add_migration(source="Island1", dest="Island2", rate=0.01)
        self.b.add_migration(source="Island2", dest="Island1", rate=0.02)
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 2)

    def test_demog_attributes(self):
        self.assertTrue(
            self.demog.metadata["burnin_time"]
            == sum(self.demog.initial_sizes_list) * 2
        )


@run_model_round_trip
class TestIslandModelRateChange(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="island", time_units="generations")
        self.b.add_deme(name="Island1", epochs=[
                        dict(start_size=100, end_time=0)])
        self.b.add_deme(name="Island2", epochs=[
                        dict(start_size=200, end_time=0)])
        self.b.add_migration(
            source="Island1", dest="Island2", rate=0.01, end_time=500)
        self.b.add_migration(source="Island2", dest="Island1", rate=0.02)
        self.b.add_migration(
            source="Island1", dest="Island2", rate=0.05, start_time=500
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 2)

    def test_burnin_time(self):
        self.assertTrue(
            self.demog.metadata["burnin_time"]
            == sum(self.demog.initial_sizes_list) * 2
        )

    def test_total_sim_length(self):
        self.assertEqual(
            self.demog.total_simulation_length,
            self.demog.metadata["burnin_time"] + 500,
        )


@run_model_round_trip
class TestTwoPopMerger(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(
            description="split then merger", time_units="generations"
        )
        self.b.add_deme(name="Ancestral", epochs=[
                        dict(start_size=1000, end_time=1000)])
        self.b.add_deme(
            name="Parent1",
            epochs=[dict(start_size=500, end_time=500)],
            ancestors=["Ancestral"],
        )
        self.b.add_deme(
            name="Parent2",
            epochs=[dict(start_size=500, end_time=500)],
            ancestors=["Ancestral"],
        )
        self.b.add_deme(
            name="Child",
            epochs=[dict(start_size=1000, end_time=0)],
            ancestors=["Parent1", "Parent2"],
            proportions=[0.5, 0.5],
            start_time=500,
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 1)

    def test_total_sim_length(self):
        self.assertEqual(
            self.demog.total_simulation_length,
            self.demog.metadata["burnin_time"] + 1000,
        )


@run_model_round_trip
class TestFourWayMerger(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(
            description="split then merger", time_units="generations"
        )
        self.b.add_deme(name="Ancestral", epochs=[
                        dict(start_size=1000, end_time=1000)])
        self.b.add_deme(
            name="A",
            epochs=[dict(start_size=500, end_time=700)],
            ancestors=["Ancestral"],
        )
        self.b.add_deme(
            name="B",
            epochs=[dict(start_size=500, end_time=500)],
            ancestors=["Ancestral"],
        )
        self.b.add_deme(
            name="Parent1", epochs=[dict(start_size=200, end_time=100)], ancestors=["A"]
        )
        self.b.add_deme(
            name="Parent2", epochs=[dict(start_size=300, end_time=100)], ancestors=["A"]
        )
        self.b.add_deme(
            name="Parent3", epochs=[dict(start_size=100, end_time=100)], ancestors=["B"]
        )
        self.b.add_deme(
            name="Parent4", epochs=[dict(start_size=400, end_time=100)], ancestors=["B"]
        )
        self.b.add_deme(
            name="Child",
            epochs=[dict(start_size=1000, end_time=0)],
            ancestors=["Parent1", "Parent2", "Parent3", "Parent4"],
            proportions=[0.2, 0.3, 0.1, 0.4],
            start_time=100,
        )
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 1)

    def test_total_sim_length(self):
        self.assertEqual(
            self.demog.total_simulation_length,
            self.demog.metadata["burnin_time"] + 1000,
        )


@run_model_round_trip
class TestPulseMigration(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.b = demes.Builder(description="pulse", time_units="generations")
        self.b.add_deme(name="deme1", epochs=[
                        dict(start_size=100, end_time=0)])
        self.b.add_deme(name="deme2", epochs=[
                        dict(start_size=100, end_time=0)])
        self.b.add_pulse(sources=["deme1"], dest="deme2",
                         time=100, proportions=[0.2])
        self.g = self.b.resolve()
        self.demog = fwdpy11.discrete_demography.from_demes(self.g, 1)

    def test_total_sim_length(self):
        self.assertEqual(
            self.demog.total_simulation_length,
            self.demog.metadata["burnin_time"] + 100,
        )


def evolve_demes_model(demog) -> fwdpy11.DiploidPopulation:
    pdict = {
        "rates": (0.0, 0.0, 0.0),
        "gvalue": fwdpy11.Multiplicative(2.0),
        "demography": demog,
        "simlen": demog.total_simulation_length,
    }
    initial_sizes = demog.initial_sizes_list
    params = fwdpy11.ModelParams(**pdict)
    rng = fwdpy11.GSLrng(100)
    pop = fwdpy11.DiploidPopulation(initial_sizes, 1.0)
    fwdpy11.evolvets(rng, pop, params, 100)
    return pop


def three_way_continuous_migration(
    goes_extinct: typing.Optional[typing.List[str]] = None,
):
    b = demes.Builder(description="many migrations", time_units="generations")
    for deme in ["A", "B", "C"]:
        if goes_extinct is not None and deme in goes_extinct:
            b.add_deme(name=deme, epochs=[dict(start_size=100, end_time=10)])
        else:
            b.add_deme(name=deme, epochs=[dict(start_size=100, end_time=0)])
    b.add_migration(demes=["A", "B", "C"], rate=0.1)
    g = b.resolve()
    return g


def three_way_continuous_migration_pairwise(
    goes_extinct: typing.Optional[typing.List[str]] = None,
):
    b = demes.Builder(description="many migrations", time_units="generations")
    for deme in ["A", "B", "C"]:
        if goes_extinct is not None and deme in goes_extinct:
            b.add_deme(name=deme, epochs=[dict(start_size=100, end_time=10)])
        else:
            b.add_deme(name=deme, epochs=[dict(start_size=100, end_time=0)])
    b.add_migration(demes=["A", "B"], rate=0.1)
    b.add_migration(demes=["A", "C"], rate=0.1)
    b.add_migration(demes=["B", "C"], rate=0.1)
    g = b.resolve()
    return g


@pytest.mark.parametrize(
    "graph",
    [
        three_way_continuous_migration(),
        three_way_continuous_migration(["A"]),
        three_way_continuous_migration(["B"]),
        three_way_continuous_migration(["C"]),
        three_way_continuous_migration(["A", "B"]),
        three_way_continuous_migration(["A", "C"]),
        three_way_continuous_migration(["B", "C"]),
        three_way_continuous_migration_pairwise(),
        three_way_continuous_migration_pairwise(["A"]),
        three_way_continuous_migration_pairwise(["B"]),
        three_way_continuous_migration_pairwise(["C"]),
        three_way_continuous_migration_pairwise(["A", "B"]),
        three_way_continuous_migration_pairwise(["A", "C"]),
        three_way_continuous_migration_pairwise(["B", "C"]),
    ],
)
def test_three_way_continuous_migration_pairwise(graph):
    _ = fwdpy11.discrete_demography.from_demes(graph, 1)


@pytest.mark.parametrize(
    "graph",
    [
        three_way_continuous_migration(),
        three_way_continuous_migration(["A"]),
        three_way_continuous_migration(["B"]),
        three_way_continuous_migration(["C"]),
        three_way_continuous_migration(["A", "B"]),
        three_way_continuous_migration(["A", "C"]),
        three_way_continuous_migration(["B", "C"]),
        three_way_continuous_migration_pairwise(),
        three_way_continuous_migration_pairwise(["A"]),
        three_way_continuous_migration_pairwise(["B"]),
        three_way_continuous_migration_pairwise(["C"]),
        three_way_continuous_migration_pairwise(["A", "B"]),
        three_way_continuous_migration_pairwise(["A", "C"]),
        three_way_continuous_migration_pairwise(["B", "C"]),
    ],
)
def test_evolve_three_way_continuous_migration_pairwise(graph):
    demog = fwdpy11.discrete_demography.from_demes(graph, 1)
    pop = evolve_demes_model(demog)
    assert pop.generation == demog.total_simulation_length


def multiple_migrations_delayed():
    b = demes.Builder(description="many migrations", time_units="generations")
    b.add_deme(name="A", epochs=[dict(start_size=100, end_time=0)])
    b.add_deme(name="B", epochs=[dict(start_size=100, end_time=0)])
    b.add_deme(name="C", epochs=[dict(start_size=100, end_time=0)])
    b.add_migration(demes=["A", "B"], rate=0.1)
    b.add_migration(demes=["B", "C"], rate=0.1)
    b.add_migration(demes=["A", "C"], rate=0.1, start_time=100)
    g = b.resolve()
    return g


@pytest.mark.parametrize("graph", [multiple_migrations_delayed()])
def test_multiple_migrations_delayed(graph):
    _ = fwdpy11.discrete_demography.from_demes(graph, 1)


def splits_with_migrations():
    b = demes.Builder(description="splits with migration",
                      time_units="generations")
    b.add_deme(name="A", epochs=[dict(start_size=100, end_time=100)])
    b.add_deme(name="B", ancestors=["A"], epochs=[
               dict(start_size=100, end_time=0)])
    b.add_deme(name="C", ancestors=["A"], epochs=[
               dict(start_size=100, end_time=50)])
    b.add_deme(name="D", ancestors=["C"], epochs=[
               dict(start_size=100, end_time=0)])
    b.add_deme(name="E", ancestors=["C"], epochs=[
               dict(start_size=100, end_time=0)])
    b.add_migration(demes=["B", "C"], rate=0.1)
    b.add_migration(demes=["B", "D"], rate=0.1)
    b.add_migration(demes=["B", "E"], rate=0.1)
    b.add_migration(demes=["D", "E"], rate=0.1)
    g = b.resolve()
    return g


@pytest.mark.parametrize("graph", [splits_with_migrations()])
def test_splits_with_migrations(graph):
    _ = fwdpy11.discrete_demography.from_demes(graph, 1)


def yaml_migration_1():
    return """
description: test model
time_units: generations
demes:
  - name: A
    epochs:
      - end_time: 200
        start_size: 100
  - name: B
    ancestors: [A]
    epochs:
      - end_time: 100
        start_size: 100
  - name: C
    ancestors: [A]
    epochs:
    - end_time: 20
      start_size: 100
  - name: D
    ancestors: [B]
    epochs:
      - end_time: 2
        start_size: 100
      - end_time: 0
        start_size: 10
  - name: E
    ancestors: [B]
    epochs:
      - end_time: 2
        start_size: 100
      - end_time: 0
        start_size: 30
migrations:
  - demes: [C, D]
    rate: 6.228e-5
  - demes: [D, E]
    rate: 4.14e-5
"""


def yaml_migration_2():
    """
    Same model as previous yaml with deme name D changed to F to switch sort order.
    """
    return """description: test model
time_units: generations
demes:
  - name: A
    epochs:
      - end_time: 200
        start_size: 100
  - name: B
    ancestors: [A]
    epochs:
      - end_time: 100
        start_size: 100
  - name: C
    ancestors: [A]
    epochs:
    - end_time: 20
      start_size: 100
  - name: F
    ancestors: [B]
    epochs:
      - end_time: 2
        start_size: 100
      - end_time: 0
        start_size: 10
  - name: E
    ancestors: [B]
    epochs:
      - end_time: 2
        start_size: 100
      - end_time: 0
        start_size: 30
migrations:
  - demes: [C, F]
    rate: 6.228e-5
  - demes: [F, E]
    rate: 4.14e-5
"""


@pytest.mark.parametrize("data", [yaml_migration_1(), yaml_migration_2()])
def test_yamls_with_migration(data):
    g = demes.loads(data)
    _ = fwdpy11.discrete_demography.from_demes(g, 1)


def test_split_model_population_size_history(two_deme_split_with_ancestral_size_change):
    """
    This is a detailed test of the complete size history
    of a model that we run all the way through.
    """
    model = fwdpy11.discrete_demography.from_demes(
        two_deme_split_with_ancestral_size_change, burnin=1
    )

    @dataclass
    class DemeSizeAtTime:
        when: int
        size: int

    class DemeSizes(object):
        def __init__(self):
            self.sizes = dict()

        def __call__(self, pop, _):
            for key, value in pop.deme_sizes(as_dict=True).items():
                if key not in self.sizes:
                    self.sizes[key] = [DemeSizeAtTime(
                        when=pop.generation, size=value)]
                else:
                    self.sizes[key].append(
                        DemeSizeAtTime(when=pop.generation, size=value)
                    )

    pdict = {
        "gvalue": fwdpy11.Multiplicative(2.0),
        "rates": (0, 0, 0),
        "demography": model,
        "simlen": model.total_simulation_length,
    }
    params = fwdpy11.ModelParams(**pdict)
    pop = fwdpy11.DiploidPopulation(100, 1.0)
    rng = fwdpy11.GSLrng(90210)
    recorder = DemeSizes()
    fwdpy11.evolvets(rng, pop, params, 100, recorder=recorder)

    # The ancestral deme exists until generation 110,
    # and we only see offspring from birth time 1 on.
    assert [i.when for i in recorder.sizes[0]] == [i for i in range(1, 111)]
    # The daughter demes are seen from 110 till the end
    for deme in [1, 2]:
        assert [i.when for i in recorder.sizes[deme]] == [
            i for i in range(111, model.total_simulation_length + 1)
        ]
    # initial daughter deme sizes
    g1 = 2 ** (1 / 10) - 1
    g2 = 4 ** (1 / 10) - 1
    assert recorder.sizes[1][0].size == int(np.rint(250 * (1 + g1)))
    assert recorder.sizes[2][0].size == int(np.rint(50 * (1 + g2)))
    # final daughter deme sizes
    assert recorder.sizes[1][-1].size == 500
    assert recorder.sizes[2][-1].size == 200

    # At generation 100, the ancestral pop size changed from 100
    # to 200
    for i in recorder.sizes[0]:
        if i.when < 101:
            assert i.size == 100, f"{i}"
        else:
            assert i.size == 200, f"{i}"


@pytest.mark.parametrize("when", [i for i in range(75, 120)])
def test_evolve_population_in_two_stages(
    when, two_deme_split_with_ancestral_size_change
):
    model = fwdpy11.discrete_demography.from_demes(
        two_deme_split_with_ancestral_size_change, burnin=1
    )
    pdict = {
        "gvalue": fwdpy11.Multiplicative(2.0),
        "rates": (0, 0, 0),
        "demography": model,
        "simlen": when,
    }
    params = fwdpy11.ModelParams(**pdict)
    pop = fwdpy11.DiploidPopulation(100, 1.0)
    rng = fwdpy11.GSLrng(90210)
    fwdpy11.evolvets(rng, pop, params, 100)

    pdict["simlen"] = model.total_simulation_length - when
    params = fwdpy11.ModelParams(**pdict)

    fwdpy11.evolvets(rng, pop, params, 100)

    counts = np.unique(np.array(pop.diploid_metadata)
                       ["deme"], return_counts=True)
    assert counts[1][0] == 500
    assert counts[1][1] == 200


@pytest.mark.parametrize("when", [i for i in range(75, 120)])
def test_evolve_population_in_two_stages_with_deepcopy(
    when, two_deme_split_with_ancestral_size_change
):
    model = fwdpy11.discrete_demography.from_demes(
        two_deme_split_with_ancestral_size_change, burnin=1
    )
    pdict = {
        "gvalue": fwdpy11.Multiplicative(2.0),
        "rates": (0, 0, 0),
        "demography": model,
        "simlen": when,
    }
    params = fwdpy11.ModelParams(**pdict)
    pop = fwdpy11.DiploidPopulation(100, 1.0)
    rng = fwdpy11.GSLrng(90210)
    fwdpy11.evolvets(rng, pop, params, 100)

    pdict2 = copy.deepcopy(pdict)
    pdict2["simlen"] = model.total_simulation_length - when
    params2 = fwdpy11.ModelParams(**pdict2)

    fwdpy11.evolvets(rng, pop, params2, 100)

    counts = np.unique(np.array(pop.diploid_metadata)
                       ["deme"], return_counts=True)
    assert counts[1][0] == 500, f"{counts}"
    assert counts[1][1] == 200, f"{counts}"


@pytest.mark.parametrize("burnin", [1, 0])
def test_evolve_demes_model_starting_with_two_pops_and_no_ancestry(
    burnin,
    start_demes_model_with_two_pops,
):
    model = fwdpy11.discrete_demography.from_demes(
        start_demes_model_with_two_pops, burnin=burnin
    )
    pdict = {
        "gvalue": fwdpy11.Multiplicative(2.0),
        "rates": (0, 0, 0),
        "demography": model,
        "simlen": model.total_simulation_length,
    }
    params = fwdpy11.ModelParams(**pdict)
    initial_sizes = model.initial_sizes_list
    pop = fwdpy11.DiploidPopulation(initial_sizes, 1.0)
    for key, value in pop.deme_sizes(as_dict=True).items():
        if key == 0:
            assert value == 100
        elif key == 1:
            assert value == 75
        else:
            raise RuntimeError("unexpected key")
    rng = fwdpy11.GSLrng(101)
    fwdpy11.evolvets(rng, pop, params, 100)
    for key, value in pop.deme_sizes(as_dict=True).items():
        if key == 0:
            assert value == 200
        elif key == 1:
            assert value == 300
        else:
            raise RuntimeError("unexpected key")


def test_four_deme_split_model():
    """
    Tests primarily for lack of temporal
    overlap b/w ancestral/derived demes.
    See GitHub issue #814
    """
    yaml = """time_units: generations
demes:
- name: A
  epochs:
    - {end_time: 45, start_size: 10}
- name: B
  ancestors: [A]
  epochs:
    - {end_time: 30, start_size: 20}
- name: C
  ancestors: [B]
  epochs:
    - {end_time: 15, start_size: 30}
- name: D
  ancestors: [C]
  epochs:
    - {end_time: 0, start_size: 40}
"""
    burnin = 1
    g = demes.loads(yaml)
    model = fwdpy11.discrete_demography.from_demes(g, burnin=burnin)

    @dataclass
    class DemeSizeAtTime:
        when: int
        size: int

    class DemeSizes(object):
        def __init__(self):
            self.sizes = dict()

        def __call__(self, pop, _):
            deme_sizes = pop.deme_sizes()
            assert len(deme_sizes[0]) == 1
            for key, value in pop.deme_sizes(as_dict=True).items():
                if key not in self.sizes:
                    self.sizes[key] = [DemeSizeAtTime(
                        when=pop.generation, size=value)]
                else:
                    self.sizes[key].append(
                        DemeSizeAtTime(when=pop.generation, size=value)
                    )

    pdict = {
        "gvalue": fwdpy11.Multiplicative(2.0),
        "rates": (0, 0, 0),
        "demography": model,
        "simlen": model.total_simulation_length,
    }
    params = fwdpy11.ModelParams(**pdict)
    initial_size = g["A"].epochs[0].start_size
    pop = fwdpy11.DiploidPopulation(initial_size, 1.0)
    rng = fwdpy11.GSLrng(90210)
    recorder = DemeSizes()
    fwdpy11.evolvets(rng, pop, params, 100, recorder=recorder)

    for deme in recorder.sizes:
        if deme > 0:
            assert len(recorder.sizes[deme]) == 15
        else:
            assert len(recorder.sizes[deme]) == burnin * initial_size


def no_demography_no_burnin():
    yaml = """time_units: generations
demes:
- name: A
  epochs:
  - {end_time: 0, start_size: 10}
"""
    burnin = 0
    g = demes.loads(yaml)
    model = fwdpy11.discrete_demography.from_demes(g, burnin=burnin)

    @dataclass
    class DemeSizeAtTime:
        when: int
        size: int

    class DemeSizes(object):
        def __init__(self):
            self.sizes = dict()

        def __call__(self, pop, _):
            deme_sizes = pop.deme_sizes()
            assert len(deme_sizes[0]) == 1
            for key, value in pop.deme_sizes(as_dict=True).items():
                if key not in self.sizes:
                    self.sizes[key] = [DemeSizeAtTime(
                        when=pop.generation, size=value)]
                else:
                    self.sizes[key].append(
                        DemeSizeAtTime(when=pop.generation, size=value)
                    )

    pdict = {
        "gvalue": fwdpy11.Multiplicative(2.0),
        "rates": (0, 0, 0),
        "demography": model,
        "simlen": model.total_simulation_length,
    }
    params = fwdpy11.ModelParams(**pdict)
    initial_size = g["A"].epochs[0].start_size
    pop = fwdpy11.DiploidPopulation(initial_size, 1.0)
    rng = fwdpy11.GSLrng(90210)
    recorder = DemeSizes()
    fwdpy11.evolvets(rng, pop, params, 100, recorder=recorder)

    assert len(recorder.sizes[0]) == 1
    assert recorder.sizes[0][0].when == 1
    assert recorder.sizes[0][0].size == 10


# The following fixtures are demes models
# with an event that should affect the first generation
# of the forward sim


@dataclass
class PedigreeRow:
    when: int
    deme: int
    label: int
    parents: list
    parent_deme: int

    def is_migrant(self) -> bool:
        return self.deme != self.parent_deme

    def is_selfed(self) -> bool:
        return self.parents[0] == self.parents[1]


@dataclass
class CrudePedigree:
    records: typing.List[PedigreeRow]
    individual_to_deme: list
    individual_record: list

    def __call__(self, pop):
        temp, temp2 = [], []
        for i, md in enumerate(pop.diploid_metadata):
            assert i == md.label

            row = PedigreeRow(pop.generation, md.deme, i, [-1, -1], md.deme)

            if len(self.individual_to_deme) > 0:
                row.parents = [self.individual_record[i] for i in md.parents]
                assert (
                    self.individual_to_deme[md.parents[0]]
                    == self.individual_to_deme[md.parents[1]]
                )
                row.parent_deme = self.individual_to_deme[md.parents[0]]
            else:
                assert pop.generation == 0

            self.records.append(row)
            temp.append(md.deme)
            temp2.append(len(self.records) - 1)

        self.individual_to_deme = temp
        self.individual_record = temp2


@dataclass
class TrackedInfo:
    when: int
    deme: int
    nselfed: int
    demesize: int
    popsize: int


@dataclass
class InfoTracker:
    data: typing.List[TrackedInfo]
    pedigree: CrudePedigree

    def __call__(self, pop, _):
        deme_sizes = pop.deme_sizes(as_dict=True)
        selfed_by_deme = {}
        for i in deme_sizes.keys():
            selfed_by_deme[i] = 0

        for md in pop.diploid_metadata:
            if md.parents[0] == md.parents[1]:
                selfed_by_deme[md.deme] += 1

        for deme, size in deme_sizes.items():
            self.data.append(
                TrackedInfo(pop.generation, deme,
                            selfed_by_deme[deme], size, pop.N)
            )

        self.pedigree(pop)


def build_tracker() -> InfoTracker:
    return InfoTracker([], CrudePedigree([], [], []))


def base_demes_model():
    yaml = """
    time_units: generations
    demes:
    - name: ancestor
      epochs:
        - {end_time: 2, start_size: 100}"""
    return yaml


def set_selfing_rate_generation_1():
    yaml = """
    - name: derived
      ancestors: [ancestor]
      epochs:
        - {end_time: 1, selfing_rate: 0.0, start_size: 100}
        - {end_time: 0, selfing_rate: 1.0, start_size: 100}
    """

    def validate(tracker: InfoTracker):
        for i in tracker.data:
            if i.deme == 0:
                assert i.when == 0
            else:
                assert i.when > 0
                if i.when == 2:
                    assert i.nselfed == i.demesize
                else:
                    # Not rigorous, but the probability
                    # of this being true is quite small
                    assert i.nselfed < i.demesize

        for i in tracker.pedigree.records:
            if i.deme == 1 and i.when == 1:
                assert i.parent_deme == 0

    return base_demes_model() + yaml, build_tracker(), validate


def set_deme_size_generation_1():
    yaml = """
    - name: derived
      ancestors: [ancestor]
      epochs:
        - {end_time: 0, start_size: 66}
    """

    def validate(tracker: InfoTracker):
        for i in tracker.data:
            if i.deme < 1:
                assert i.demesize == 100
                assert i.when == 0
            else:
                assert i.demesize == 66
                assert i.when > 0

    return base_demes_model() + yaml, build_tracker(), validate


def set_growth_rate_generation_1():
    yaml = """
    - name: derived
      ancestors: [ancestor]
      epochs:
        - {end_time: 0, start_size: 50, end_size: 250}
    """

    def validate(tracker: InfoTracker):
        assert tracker.data[-1].demesize == 250
        # Check the initial size of deme 1
        g = 5 ** (1 / 2) - 1
        for i in tracker.data:
            if i.deme == 1:
                assert i.when > 0
                assert i.demesize == np.rint(50 * (1 + g))
                break
            else:
                assert i.when == 0

    return base_demes_model() + yaml, build_tracker(), validate


def simple_split_generation_1():
    yaml = """
    - name: derived0
      ancestors: [ancestor]
      epochs:
        - {end_time: 0, start_size: 50}
    - name: derived1
      ancestors: [ancestor]
      epochs:
        - {end_time: 0, start_size: 50}
    """

    def validate(tracker: InfoTracker):
        for i in tracker.data:
            if i.deme < 1:
                assert i.when < 1
                assert i.demesize == 100
            else:
                assert i.when > 0
                assert i.demesize == 50

    return base_demes_model() + yaml, build_tracker(), validate


def simple_split_with_migration_generation_1():
    yaml = """
    - name: derived0
      ancestors: [ancestor]
      epochs:
        - {end_time: 0, start_size: 50}
    - name: derived1
      ancestors: [ancestor]
      epochs:
        - {end_time: 0, start_size: 50}
    migrations:
      - source: derived0
        dest: derived1
        rate: 1.0
    """

    def validate(tracker: InfoTracker):
        for i in tracker.data:
            if i.deme < 1:
                assert i.when < 1
                assert i.demesize == 100
            else:
                assert i.when > 0
                assert i.demesize == 50

        for i in tracker.pedigree.records:
            if i.deme == 2:
                assert i.is_migrant()
                if i.when > 1:
                    assert i.parent_deme == 1
                else:
                    assert i.parent_deme == 0
            elif i.deme == 1:
                if i.when == 2:
                    assert not i.is_migrant()
                else:
                    assert i.is_migrant()

        # All parents of demes 1 and 2 must be from 0 at generation 1
        for i in tracker.pedigree.records:
            if i.when == 1:
                assert i.deme == 1 or i.deme == 2
                assert i.parent_deme == 0

        deme1parents, deme2parents = [], []
        for i in tracker.pedigree.records:
            if i.when == 2:
                if i.deme == 1:
                    deme1parents.append(i.parent_deme)
                elif i.deme == 2:
                    deme2parents.append(i.parent_deme)

        assert all([i == 1 for i in deme1parents])
        assert not any([i == 2 for i in deme2parents])

    return base_demes_model() + yaml, build_tracker(), validate


def simple_split_with_growth_generation_1():
    yaml = """
    - name: derived0
      ancestors: [ancestor]
      epochs:
        - {end_time: 0, start_size: 50, end_size: 100}
    - name: derived1
      ancestors: [ancestor]
      epochs:
        - {end_time: 0, start_size: 75, end_size: 1000}
    """

    def validate(tracker: InfoTracker):
        for i in tracker.data:
            if i.deme < 1:
                assert i.when == 0
            else:
                assert i.when > 0

        for i in tracker.data[::-1]:
            if i.deme == 1:
                assert i.demesize == 100
                break

        for i in tracker.data[::-1]:
            if i.deme == 2:
                assert i.demesize == 1000
                break

        g = (100 / 50) ** (1 / 2) - 1
        for i in tracker.data:
            if i.deme == 1:
                assert i.demesize == np.rint(50 * (1 + g))
                break

        g = (1000 / 75) ** (1 / 2) - 1
        for i in tracker.data:
            if i.deme == 2:
                assert i.demesize == np.rint(75 * (1 + g))
                break

        # All parents of demes 1 and 2 must be from 0 at generation 1
        for i in tracker.pedigree.records:
            if i.when == 1:
                assert i.deme == 1 or i.deme == 2
                assert i.parent_deme == 0

        deme1parents, deme2parents = [], []
        for i in tracker.pedigree.records:
            if i.when == 2:
                if i.deme == 1:
                    deme1parents.append(i.parent_deme)
                elif i.deme == 2:
                    deme2parents.append(i.parent_deme)

        assert all([i == 1 for i in deme1parents])
        assert all([i == 2 for i in deme2parents])

    return base_demes_model() + yaml, build_tracker(), validate


def simple_split_with_growth_and_migration_generation_1():
    yaml = """
    - name: derived0
      ancestors: [ancestor]
      epochs:
        - {end_time: 0, start_size: 50, end_size: 100}
    - name: derived1
      ancestors: [ancestor]
      epochs:
        - {end_time: 0, start_size: 75, end_size: 1000}
    migrations:
      - source: derived0
        dest: derived1
        rate: 0.01
    """

    def validate(tracker: InfoTracker):
        for i in tracker.data:
            if i.deme < 1:
                assert i.when == 0
            else:
                assert i.when > 0

        for i in tracker.data[::-1]:
            if i.deme == 1:
                assert i.demesize == 100
                break

        for i in tracker.data[::-1]:
            if i.deme == 2:
                assert i.when == 2
                assert i.demesize == 1000
                break

        g = (100 / 50) ** (1 / 2) - 1
        for i in tracker.data:
            if i.deme == 1:
                assert i.when == 1
                assert i.demesize == np.rint(50 * (1 + g))
                break

        g = (1000 / 75) ** (1 / 2) - 1
        for i in tracker.data:
            if i.deme == 2:
                assert i.demesize == np.rint(75 * (1 + g))
                break

        # All parents of demes 1 and 2 must be from 0 at generation 1
        for i in tracker.pedigree.records:
            if i.when == 1:
                assert i.deme == 1 or i.deme == 2
                assert i.parent_deme == 0

        deme1parents, deme2parents = [], []
        for i in tracker.pedigree.records:
            if i.when == 2:
                if i.deme == 1:
                    deme1parents.append(i.parent_deme)
                elif i.deme == 2:
                    deme2parents.append(i.parent_deme)

        assert all([i == 1 for i in deme1parents])
        assert any([i == 2 for i in deme2parents])

    return base_demes_model() + yaml, build_tracker(), validate


def one_generation_hop_generation_1():
    yaml = """
    - name: transient
      ancestors: [ancestor]
      epochs:
        - {end_time: 1, start_size: 100}
    - name: final
      ancestors: [transient]
      epochs:
        - {end_time: 0, start_size: 200}
    """

    def validate(tracker: InfoTracker):
        demes_seen = {}
        for i in tracker.data:
            demes_seen[i.deme] = True
            if i.deme == 0:
                assert i.when == 0
            elif i.deme == 1:
                assert i.when == 1
                assert i.demesize == 100
            else:
                assert i.when == 2
                assert i.demesize == 200

        assert 0 in demes_seen
        assert 1 in demes_seen
        assert 2 in demes_seen

    return base_demes_model() + yaml, build_tracker(), validate


@pytest.mark.parametrize(
    "testdata",
    [
        set_selfing_rate_generation_1,
        set_deme_size_generation_1,
        set_growth_rate_generation_1,
        simple_split_generation_1,
        simple_split_with_migration_generation_1,
        simple_split_with_growth_generation_1,
        simple_split_with_growth_and_migration_generation_1,
        one_generation_hop_generation_1,
    ],
)
@pytest.mark.parametrize("burnin", [0, 1])
def test_events_in_generation_one_following_demes_import(testdata, burnin):
    (model, recorder, validate) = testdata()
    g = demes.loads(model)
    demog = fwdpy11.discrete_demography.from_demes(g, burnin=0)
    pdict = {
        "gvalue": fwdpy11.Multiplicative(2.0),
        "rates": (0, 0, 0),
        "demography": demog,
        "simlen": demog.total_simulation_length,
    }
    params = fwdpy11.ModelParams(**pdict)

    initial_sizes = demog.initial_sizes_list
    pop = fwdpy11.DiploidPopulation(initial_sizes, 1.0)
    recorder(pop, None)
    rng = fwdpy11.GSLrng(90210)
    fwdpy11.evolvets(rng, pop, params, 100, recorder=recorder)

    validate(recorder)


@pytest.mark.parametrize(
    "testdata",
    [
        set_selfing_rate_generation_1,
        set_deme_size_generation_1,
        set_growth_rate_generation_1,
        simple_split_generation_1,
        simple_split_with_migration_generation_1,
        simple_split_with_growth_generation_1,
        simple_split_with_growth_and_migration_generation_1,
        one_generation_hop_generation_1,
    ],
)
@pytest.mark.parametrize("burnin", [0, 1])
def test_events_in_generation_one_following_demes_import_start_stop(testdata, burnin):
    (model, recorder, validate) = testdata()
    g = demes.loads(model)
    demog = fwdpy11.discrete_demography.from_demes(g, burnin=0)
    pdict = {
        "gvalue": fwdpy11.Multiplicative(2.0),
        "rates": (0, 0, 0),
        "demography": demog,
        "simlen": 1,
    }
    params = fwdpy11.ModelParams(**pdict)

    initial_sizes = demog.initial_sizes_list
    pop = fwdpy11.DiploidPopulation(initial_sizes, 1.0)
    recorder(pop, None)
    rng = fwdpy11.GSLrng(90210)
    fwdpy11.evolvets(rng, pop, params, 100, recorder=recorder)

    pdict["simlen"] = demog.total_simulation_length - 1
    fwdpy11.evolvets(
        rng, pop, params, 100, recorder=recorder,
    )

    validate(recorder)


@pytest.fixture
def multiple_pulse_source_setup():
    b = demes.Builder(description="pulse", time_units="generations")
    b.add_deme(name="deme0", epochs=[dict(start_size=100, end_time=0)])
    b.add_deme(name="deme1", epochs=[dict(start_size=100, end_time=0)])
    b.add_deme(name="deme2", epochs=[dict(start_size=100, end_time=0)])
    b.add_pulse(
        sources=["deme0", "deme1"], dest="deme2", time=100, proportions=[0.2, 0.25]
    )
    g = b.resolve()
    demog = fwdpy11.discrete_demography.from_demes(g, 10)

    return demog


def test_multiple_pulse_source(multiple_pulse_source_setup):
    demog = multiple_pulse_source_setup
    assert (
        demog.total_simulation_length == demog.metadata["burnin_time"] + 100
    )


@pytest.fixture
def two_demes_migration_rate_changes_setup():
    b = demes.Builder(time_units="generations")
    b.add_deme(name="A", epochs=[dict(start_size=100)])
    b.add_deme(name="B", epochs=[dict(start_size=100)])
    b.add_migration(demes=["A", "B"], rate=0.0, end_time=30)
    b.add_migration(demes=["A", "B"], rate=0.5, start_time=30, end_time=20)
    b.add_migration(demes=["A", "B"], rate=0.0, start_time=20, end_time=10)
    b.add_migration(demes=["A", "B"], rate=0.5, start_time=10)
    g = b.resolve()
    return g


def test_two_demes_migration_rate_changes(two_demes_migration_rate_changes_setup):
    # test to check that the last 10 generation have the high migration rates
    # not 9 generations, and not 11 generations

    @dataclass
    class ParentDemesAtTime:
        when: int
        parents: list

    class ParentDemes(object):
        def __init__(self):
            self.parents = dict()

        def __call__(self, pop, _):
            deme_indexes, deme_sizes = pop.deme_sizes()
            for deme_idx in deme_indexes:
                if deme_idx not in self.parents:
                    self.parents[deme_idx] = [
                        ParentDemesAtTime(when=pop.generation, parents=[0, 0])
                    ]
                else:
                    self.parents[deme_idx].append(
                        ParentDemesAtTime(when=pop.generation, parents=[0, 0])
                    )
            for metadata in pop.diploid_metadata:
                parents = metadata.parents
                for parent in parents:
                    self.parents[metadata.deme][-1].parents[
                        pop.diploid_metadata[parent].deme
                    ] += 1

    demog = fwdpy11.discrete_demography.from_demes(
        two_demes_migration_rate_changes_setup, 1
    )

    pdict = {
        "gvalue": fwdpy11.Multiplicative(2.0),
        "rates": (0, 0, 0),
        "demography": demog,
        "simlen": demog.total_simulation_length,
    }
    params = fwdpy11.ModelParams(**pdict)
    initial_sizes = demog.initial_sizes_list
    pop = fwdpy11.DiploidPopulation([initial_sizes[0], initial_sizes[1]], 1.0)
    rng = fwdpy11.GSLrng(90210)
    recorder = ParentDemes()
    fwdpy11.evolvets(rng, pop, params, 100, recorder=recorder)
    for ii in range(1, 11):
        assert (
            recorder.parents[0][-ii].parents[0] > 0
            and recorder.parents[0][-ii].parents[1] > 0
        )
        assert (
            recorder.parents[0][-ii].parents[0] > 0
            and recorder.parents[0][-ii].parents[1] > 0
        )
        assert (
            recorder.parents[0][-ii - 20].parents[0] > 0
            and recorder.parents[0][-ii - 20].parents[1] > 0
        )
        assert (
            recorder.parents[0][-ii - 20].parents[0] > 0
            and recorder.parents[0][-ii - 20].parents[1] > 0
        )
        assert recorder.parents[0][-ii - 10].parents[0] == 200
        assert recorder.parents[1][-ii - 10].parents[1] == 200
        assert recorder.parents[0][-ii - 30].parents[0] == 200
        assert recorder.parents[1][-ii - 30].parents[1] == 200


def test_split_with_existence_n_way_migration():
    b = demes.Builder(time_units="generations")
    b.add_deme(name="anc", epochs=[dict(start_size=100, end_time=200)])
    b.add_deme(name="A", ancestors=["anc"], epochs=[dict(start_size=100)])
    b.add_deme(name="B", ancestors=["anc"], epochs=[
               dict(start_size=100, end_time=100)])
    b.add_deme(name="C", ancestors=["B"], epochs=[dict(start_size=100)])
    b.add_deme(name="D", ancestors=["B"], epochs=[dict(start_size=100)])
    b.add_deme(name="E", ancestors=["B"], epochs=[dict(start_size=100)])
    b.add_migration(demes=["A", "C", "D", "E"], rate=0.1)
    resolve_and_run(b, burnin=1)


def test_non_integer_size_change():
    b = demes.Builder()
    b.add_deme("A", epochs=[{"start_size": 100}])
    b.add_deme("B", ancestors=["A"], start_time=10,
               epochs=[{"start_size": 100.5}])
    g = b.resolve()

    _ = fwdpy11.discrete_demography.from_demes(g, round_non_integer_sizes=True)


def test_non_integer_initial_epoch_size():
    b = demes.Builder()
    b.add_deme("A", epochs=[{"start_size": 100.5}])
    g = b.resolve()

    demog = fwdpy11.discrete_demography.from_demes(
        g, round_non_integer_sizes=True)

    # initialize the ancestral population
    initial_sizes = demog.initial_sizes_list
    _ = fwdpy11.DiploidPopulation(
        initial_sizes, 1)


def test_very_short_epoch():
    b = demes.Builder()
    b.add_deme("A", epochs=[{"start_size": 100}])
    b.add_deme(
        "B",
        ancestors=["A"],
        start_time=10,
        epochs=[
            {"start_size": 100, "end_time": 1e-3},
            {"start_size": 200, "end_time": 0.0},
        ],
    )
    g = b.resolve()

    with pytest.raises(ValueError):
        _ = fwdpy11.discrete_demography.from_demes(g)


def test_epoch_rounding_01():
    yaml = """
time_units: generations
demes:
- name: bad
  epochs:
  - {end_time: 1.75, start_size: 1}
  - {end_time: 1.25, start_size: 2}
  - {end_time: 0.5, start_size: 3}
  - {end_time: 0, start_size: 4}
"""
    graph = demes.loads(yaml)
    with pytest.raises(ValueError):
        _ = fwdpy11.discrete_demography.from_demes(graph, burnin=0)


def test_epoch_rounding_02():
    yaml = """
time_units: generations
demes:
- name: bad
  epochs:
  - {end_time: 1.5, start_size: 1}
  - {end_time: 0.4, start_size: 2}
  - {end_time: 0, start_size: 3}
"""
    graph = demes.loads(yaml)
    with pytest.raises(ValueError):
        _ = fwdpy11.discrete_demography.from_demes(graph, burnin=0)


def test_ambiguous_pulses():
    yaml = """
time_units: generations
demes:
 - name: A
   epochs:
    - start_size: 50
 - name: B
   epochs:
    - start_size: 50
pulses:
 - sources: [A]
   dest: B
   proportions: [0.9]
   time: 25
 - sources: [A]
   dest: B
   proportions: [0.9]
   time: 25
"""
    with pytest.warns(UserWarning):
        graph = demes.loads(yaml)
    with pytest.warns(UserWarning):
        _ = fwdpy11.discrete_demography.from_demes(graph, burnin=0)


def test_deme_sort_order():
    yaml = """
description: single deme model
time_units: generations
demes:
 - name: B
   epochs:
    - start_size: 100
 - name: A
   epochs:
    - start_size: 100
"""
    graph = demes.loads(yaml)
    model = fwdpy11.discrete_demography.from_demes(graph, burnin=0)
    assert model.deme_labels[0] == "B"
    assert model.deme_labels[1] == "A"


def test_residual_selfing_exception():
    yaml = """
description: single deme model
time_units: generations
demes:
 - name: B
   epochs:
    - start_size: 1
"""
    graph = demes.loads(yaml)
    model = fwdpy11.discrete_demography.from_demes(graph, burnin=10)
    pop = fwdpy11.DiploidPopulation([1], 1.0)
    pdict = {"recregions": [],
             "nregions": [],
             "rates": (0, 0, 0),
             "gvalue": fwdpy11.Multiplicative(2.),
             "simlen": 10,
             "demography": model,
             "allow_residual_selfing": False
             }
    params = fwdpy11.ModelParams(**pdict)
    rng = fwdpy11.GSLrng(42)
    with pytest.raises(fwdpy11.DemographyError):
        with pytest.warns(UserWarning):
            fwdpy11.evolvets(rng, pop, params, 5)



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import json
import typing

import fwdpy11.tskit_tools
import fwdpy11.tskit_tools.metadata_schema
import numpy as np
import tskit  # type: ignore


def _initializePopulationTable(
    node_view,
    population_metadata: typing.Optional[typing.Dict[int, object]],
    tc
):
    tc.populations.metadata_schema = (
        fwdpy11.tskit_tools.metadata_schema.PopulationMetadata
    )
    # Prior to 0.15.2, we just iterated over unique values.
    # However, that was incorrect (see GitHub issue 792)
    # because the possibility of demes going extinct could leave
    # nodes in the tree in a way that resulted in population IDs
    # being shifted, giving a LibraryError from tskit.
    for i in range(node_view["deme"].max() + 1):
        if population_metadata is not None and i in population_metadata:
            tc.populations.add_row(metadata=population_metadata[i])
        else:
            tc.populations.add_row(metadata={"name": "deme" + str(i)})


def _initializeIndividualTable(self, tc):
    """
    Returns node ID -> individual map
    """
    tc.individuals.metadata_schema = (
        fwdpy11.tskit_tools.metadata_schema.IndividualDiploidMetadata
    )
    # First, alive individuals:
    individal_nodes = {}
    num_ind_nodes = 0
    for i, d in enumerate(self.diploid_metadata):
        individal_nodes[2 * i] = i
        individal_nodes[2 * i + 1] = i
        num_ind_nodes += 1
        tc.individuals.add_row(
            flags=fwdpy11.tskit_tools.INDIVIDUAL_IS_ALIVE,
            metadata=fwdpy11.tskit_tools.metadata_schema.generate_individual_metadata(
                d
            ),
        )

    # Now, preserved nodes
    node_time = np.array(self.tables.nodes, copy=False)["time"]
    for i in self.ancient_sample_metadata:
        assert i.nodes[0] not in individal_nodes, "individual record error"
        assert i.nodes[1] not in individal_nodes, "individual record error"
        individal_nodes[i.nodes[0]] = num_ind_nodes
        individal_nodes[i.nodes[1]] = num_ind_nodes
        num_ind_nodes += 1
        flag = fwdpy11.tskit_tools.INDIVIDUAL_IS_PRESERVED
        if node_time[i.nodes[0]] == 0.0 and node_time[i.nodes[1]] == 0.0:
            flag |= fwdpy11.tskit_tools.INDIVIDUAL_IS_FIRST_GENERATION
        tc.individuals.add_row(
            flags=flag,
            metadata=fwdpy11.tskit_tools.metadata_schema.generate_individual_metadata(
                i
            ),
        )

    return individal_nodes


def _dump_mutation_site_and_site_tables(self,
                                        tc: tskit.TableCollection) -> None:
    mpos = np.array(
        [self.mutations[mr.key].pos for mr in self.tables.mutations])
    ancestral_state = np.zeros(
        len(self.tables.mutations), dtype=np.int8) + ord("0")
    ancestral_state_offset = np.arange(
        len(self.tables.mutations) + 1, dtype=np.uint32)
    tc.sites.set_columns(
        position=mpos,
        ancestral_state=ancestral_state,
        ancestral_state_offset=ancestral_state_offset,
    )

    tc.mutations.metadata_schema = (
        fwdpy11.tskit_tools.metadata_schema.determine_mutation_metadata_schema(
            self.mutations
        )
    )
    for m in self.tables.mutations:
        if self.mutations[m.key].g != np.iinfo(np.int32).min:
            origin_time = self.generation - self.mutations[m.key].g
        else:
            origin_time = tc.nodes.time[m.node]

        tc.mutations.add_row(
            site=m.site,
            node=m.node,
            derived_state="1",
            time=origin_time,
            metadata=fwdpy11.tskit_tools.metadata_schema.generate_mutation_metadata(
                m, self.mutations
            ),
        )


def _dump_tables_to_tskit(
    self,
    *,
    model_params=None,
    demes_graph=None,
    population_metadata=None,
    data=None,
    seed=None,
    parameters=None,
    destructive=False,
) -> tskit.TreeSequence:
    from .._fwdpy11 import gsl_version, pybind11_version

    environment = tskit.provenance.get_environment(
        extra_libs={
            "gsl": {"version": gsl_version()["gsl_version"]},
            "pybind11": {"version": pybind11_version()["pybind11_version"]},
        }
    )

    provenance = {
        "schema_version": "1.0.0",
        "software": {"name": "fwdpy11", "version": f"{fwdpy11.__version__}"},
        "environment": environment,
        "parameters": {},
    }

    if parameters is not None:
        provenance["parameters"] = parameters

    tskit.validate_provenance(provenance)

    node_view = np.array(self.tables.nodes, copy=True)
    node_view["time"] -= node_view["time"].max()
    node_view["time"][np.where(node_view["time"] != 0.0)[0]] *= -1.0
    edge_view = np.array(self.tables.edges, copy=False)

    tc = tskit.TableCollection(self.tables.genome_length)

    tc.metadata_schema = fwdpy11.tskit_tools.metadata_schema.TopLevelMetadata

    # Populate the required fields
    top_level_metadata = {"generation": self.generation}

    if model_params is not None:
        try:
            top_level_metadata["model_params"] = str(model_params.asdict())
        except Exception:
            mp = {}
            for key, value in model_params.items():
                mp[key] = str(value.asdict())
            top_level_metadata["model_params"] = mp

    if demes_graph is not None:
        top_level_metadata["demes_graph"] = demes_graph.asdict()

    if data is not None:
        top_level_metadata["data"] = data

    if seed is not None:
        if seed < 0:
            raise ValueError(f"seed must be >=0, got {seed}")
        top_level_metadata["seed"] = seed

    tc.metadata = top_level_metadata

    if destructive is True:
        self._clear_haploid_genomes()

    # We must initialize population and individual
    # tables before we can do anything else.
    # Attempting to set population to anything
    # other than -1 in an tskit.NodeTable will
    # raise an exception if the PopulationTable
    # isn't set up.
    _initializePopulationTable(node_view, population_metadata, tc)
    node_to_individual = _initializeIndividualTable(self, tc)
    individual = [-1 for i in range(len(node_view))]
    for k, v in node_to_individual.items():
        individual[k] = v
    flags = [tskit.NODE_IS_SAMPLE] * 2 * self.N + [0] * (len(node_view) - 2 * self.N)
    # Bug fixed in 0.3.1: add preserved nodes to samples list
    for i in np.array(self.ancient_sample_metadata,
                      copy=False)["nodes"].flatten():
        flags[i] = tskit.NODE_IS_SAMPLE
    tc.nodes.set_columns(
        flags=flags,
        time=node_view["time"],
        population=node_view["deme"],
        individual=individual,
    )
    if destructive is True:
        self._clear_diploid_metadata()
        self._clear_ancient_sample_metadata()
        self.tables._clear_nodes()

    _dump_mutation_site_and_site_tables(self, tc)
    if destructive is True:
        self._clear_mutations()
        self.tables._clear_sites()
        self.tables._clear_mutations()

    tc.edges.set_columns(
        left=edge_view["left"],
        right=edge_view["right"],
        parent=edge_view["parent"],
        child=edge_view["child"],
    )
    if destructive is True:
        self.tables._clear_edges()

    tc.provenances.add_row(json.dumps(provenance))

    return tc.tree_sequence()


def _add_dump_tables_to_tskit(cls):
    cls.dump_tables_to_tskit = _dump_tables_to_tskit



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#


def get_includes():
    """
    Returns absolute path to location of fwdpy11 headers
    """
    import os
    import fwdpy11

    return os.path.dirname(fwdpy11.__file__) + "/headers"


def get_fwdpp_includes():
    """
    Returns absolute path to location of the fwdpp headers
    installed along with fwdpy11.
    """
    return get_includes() + "/fwdpp"


def minimal_mako():
    """
    Returns a minimal mako header for compiling
    plugins using cppimport

    .. versionadded:: 0.1.1
    """
    rv = r"""<%
setup_pybind11(cfg)
import fwdpy11 as fp11
cfg['include_dirs'] = [fp11.get_includes(), fp11.get_fwdpp_includes()]
%>"""
    return rv



# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your meanion) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.

# Main goal is to test pickling

import unittest


class testNoNoise(unittest.TestCase):
    @classmethod
    def setUp(self):
        import fwdpy11

        self.n = fwdpy11.NoNoise()

    def testPickle(self):
        import pickle

        p = pickle.dumps(self.n, -1)
        up = pickle.loads(p)
        self.assertEqual(type(up), type(self.n))


class testNoNoisePackageAlias(unittest.TestCase):
    @classmethod
    def setUp(self):
        import fwdpy11 as fancyname

        self.n = fancyname.NoNoise()

    def testPickle(self):
        import pickle

        p = pickle.dumps(self.n, -1)
        up = pickle.loads(p)
        self.assertEqual(type(up), type(self.n))


class testGaussianNoise(unittest.TestCase):
    @classmethod
    def setUp(self):
        import fwdpy11

        self.sd = 1.0
        self.mean = 0.0
        self.n = fwdpy11.GaussianNoise(sd=self.sd, mean=self.mean)

    def testPickle(self):
        import pickle

        p = pickle.dumps(self.n, -1)
        up = pickle.loads(p)
        self.assertEqual(up.sd, self.sd)
        self.assertEqual(up.mean, self.mean)


if __name__ == "__main__":
    unittest.main()



import attr

from fwdpy11.class_decorators import (attr_add_asblack, attr_class_pickle,
                                      attr_class_to_from_dict,)


@attr_add_asblack
@attr_class_pickle
@attr_class_to_from_dict
@attr.s(
    kw_only=True, frozen=True, auto_attribs=True, repr_ns="fwdpy11"
)
class DemographicModelCitation(object):
    """
    Citation information for a demographic model

    This class has the following attributes, whose names
    are also ``kwargs`` for intitialization.  The attribute names
    also determine the order of positional arguments:

    :param DOI: The Digital Object Identifier
    :param full_citation: Something string-like giving the full citation.
    :param metadata: Any additional information that may be needed.


    .. versionadded:: 0.8.0
    """

    DOI: object
    full_citation: object
    metadata: object



#
# Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
#
# This file is part of fwdpy11.
#
# fwdpy11 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# fwdpy11 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
#

import json

import fwdpy11
import msprime
import pytest


@pytest.fixture
def pdict(request):
    pd = {
        "recregions": [],
        "sregions": [],
        "nregions": [],
        "rates": (0, 0, 0),
        "gvalue": fwdpy11.Multiplicative(2.0),
        "demography": None,
        "simlen": request.param["simlen"],
    }
    return pd


@pytest.fixture
def pop(request):
    return fwdpy11.DiploidPopulation(request.param["N"], request.param["L"])


@pytest.fixture
def inception():
    with open(__file__, "r") as f:
        contents = f.read()
    return contents


@pytest.mark.parametrize("rng", [{"seed": 666}], indirect=["rng"])
@pytest.mark.parametrize("pdict", [{"simlen": 10}], indirect=["pdict"])
@pytest.mark.parametrize("pop", [{"N": 100, "L": 1}], indirect=["pop"])
def test_metadata_roundtrip_single_sim(rng, pdict, pop):
    demography = fwdpy11.ForwardDemesGraph.tubes(pop.deme_sizes()[1],
                                                 burnin=pdict["simlen"],
                                                 burnin_is_exact=True)
    pdict["demography"] = demography
    params = fwdpy11.ModelParams(**pdict)

    r = fwdpy11.RandomAncientSamples(seed=42, samplesize=2, timepoints=[3])

    fwdpy11.evolvets(rng, pop, params, 100, r)

    ts = pop.dump_tables_to_tskit(model_params=params)

    # add neutral mutations w/no metadata
    ts = msprime.sim_mutations(ts, rate=1.0, random_seed=654321)
    # bulk decode the mutation metadata, which is all None
    mutation_md = fwdpy11.tskit_tools.decode_mutation_metadata(ts)
    assert all([i is None for i in mutation_md])

    # test index access
    num_muts = 0
    for i in range(ts.num_mutations):
        x = fwdpy11.tskit_tools.decode_mutation_metadata(ts, i)
        assert x[0] is None
        num_muts += 1
    assert num_muts == ts.num_mutations

    num_muts = 0
    for i in fwdpy11.tskit_tools.decode_mutation_metadata(
        ts, slice(0, ts.num_mutations, 2)
    ):
        assert i is None
        num_muts += 1

    assert num_muts == len([i for i in range(0, ts.num_mutations, 2)])

    assert ts.num_individuals == pop.N + 2
    first = 0
    preserved = 0
    alive = 0
    for i in ts.individuals():
        if i.flags & fwdpy11.tskit_tools.INDIVIDUAL_IS_ALIVE:
            alive += 1
            for n in i.metadata["nodes"]:
                assert ts.node(n).time == 0
        if i.flags & fwdpy11.tskit_tools.INDIVIDUAL_IS_PRESERVED:
            preserved += 1
            for n in i.metadata["nodes"]:
                assert (
                    ts.node(n).time == 0 + params.simlen
                    or ts.node(n).time == 0 + params.simlen - 3
                )

    assert alive == pop.N
    assert first == 0
    assert preserved == 2
    # Test slice/index access to individual metadata
    first = 0
    preserved = 0
    alive = 0
    for row in range(ts.num_individuals):
        i = fwdpy11.tskit_tools.decode_individual_metadata(ts, row)[0]
        assert type(i.alive) == bool
        assert type(i.preserved) == bool
        assert type(i.first_generation) == bool
        if i.alive:
            alive += 1
            for n in i.nodes:
                assert ts.node(n).time == 0
        if i.preserved:
            preserved += 1
            for n in i.nodes:
                assert (
                    ts.node(n).time == 0 + params.simlen
                    or ts.node(n).time == 0 + params.simlen - 3
                )
    assert alive == pop.N
    assert first == 0
    assert preserved == 2

    num_md = 0
    for _ in fwdpy11.tskit_tools.decode_individual_metadata(
        ts, slice(1, ts.num_individuals, 2)
    ):
        num_md += 1
    assert num_md == len([i for i in range(0, ts.num_individuals, 2)])


@pytest.mark.parametrize("rng", [{"seed": 1234}], indirect=["rng"])
@pytest.mark.parametrize("pdict", [{"simlen": 10}], indirect=["pdict"])
@pytest.mark.parametrize("pop", [{"N": 100, "L": 1}], indirect=["pop"])
def test_metadata_roundtrip_single_sim_with_first_gen_preserved(rng, pdict, pop):
    demography = fwdpy11.ForwardDemesGraph.tubes(pop.deme_sizes()[1],
                                                 burnin=pdict["simlen"],
                                                 burnin_is_exact=True)
    pdict["demography"] = demography
    params = fwdpy11.ModelParams(**pdict)

    r = fwdpy11.RandomAncientSamples(seed=42, samplesize=2, timepoints=[3])

    fwdpy11.evolvets(rng, pop, params, 100, r, preserve_first_generation=True)

    ts = pop.dump_tables_to_tskit(model_params=params)
    assert ts.num_individuals == 2 * pop.N + 2
    first = 0
    preserved = 0
    alive = 0
    for i in ts.individuals():
        if i.flags & fwdpy11.tskit_tools.INDIVIDUAL_IS_ALIVE:
            alive += 1
            for n in i.metadata["nodes"]:
                assert ts.node(n).time == 0
        if i.flags & fwdpy11.tskit_tools.INDIVIDUAL_IS_PRESERVED:
            preserved += 1
            for n in i.metadata["nodes"]:
                assert (
                    ts.node(n).time == 0 + params.simlen
                    or ts.node(n).time == 0 + params.simlen - 3
                )
        if i.flags & fwdpy11.tskit_tools.INDIVIDUAL_IS_FIRST_GENERATION:
            first += 1
            for n in i.metadata["nodes"]:
                assert ts.node(n).time == 0 + params.simlen

    assert alive == pop.N
    assert first == pop.N
    assert preserved == 2 + pop.N


@pytest.mark.parametrize("rng", [{"seed": 666}], indirect=["rng"])
@pytest.mark.parametrize("pdict", [{"simlen": 10}], indirect=["pdict"])
@pytest.mark.parametrize("pop", [{"N": 100, "L": 1}], indirect=["pop"])
def test_metadata_roundtrip_single_deme_sim_with_parameters(rng, pdict, pop, inception):
    demography = fwdpy11.ForwardDemesGraph.tubes(pop.deme_sizes()[1],
                                                 burnin=pdict["simlen"],
                                                 burnin_is_exact=True)
    pdict["demography"] = demography
    params = fwdpy11.ModelParams(**pdict)

    fwdpy11.evolvets(rng, pop, params, 100)

    ts = pop.dump_tables_to_tskit(
        parameters={"params_dict": str(params.asdict()), "script": inception}
    )

    provenance = json.loads(ts.provenance(0).record)
    params_dict = eval(provenance["parameters"]["params_dict"])

    assert params_dict == params.asdict()
    script = provenance["parameters"]["script"]

    with open(__file__, "r") as f:
        assert f.read() == script



"""
Unit tests for the Demes reference parser implementation.

Run with ``python3 -m pytest ``
"""
import pathlib
import json
import math

import jsonschema
import pytest
from ruamel.yaml import YAML
from ruamel.yaml.constructor import ConstructorError

import demes_parser as parser


def minimal_graph(num_demes=1, population_size=1):
    graph = {
        "time_units": "generations",
        "demes": [
            {"name": f"deme{j}", "epochs": [{"start_size": population_size}]}
            for j in range(num_demes)
        ],
    }
    return graph


def island_model_graph(num_demes=1, population_size=1, migration_rate=1):
    graph = {
        "time_units": "generations",
        "demes": [
            {"name": f"deme{j}", "epochs": [{"start_size": population_size}]}
            for j in range(num_demes)
        ],
        "migrations": [
            {"rate": migration_rate, "demes": [f"deme{j}" for j in range(num_demes)]}
        ],
    }
    return graph


def single_deme_graph(num_epochs=1, population_size=1):
    graph = {
        "time_units": "generations",
        "demes": [
            {
                "name": "deme0",
                "epochs": [
                    {"start_size": population_size + j, "end_time": num_epochs - j - 1}
                    for j in range(num_epochs)
                ],
            }
        ],
    }
    return graph


def single_ancestor_graph(num_demes=1, population_size=1):
    graph = {
        "time_units": "generations",
        "defaults": {"epoch": {"start_size": population_size}},
        "demes": [{"name": "ancestor", "epochs": [{"end_time": 10}]}]
        + [{"name": f"child_{j}", "ancestors": ["ancestor"]} for j in range(num_demes)],
    }
    return graph


def two_ancestor_graph(num_demes=1, population_size=1):
    graph = {
        "time_units": "generations",
        "defaults": {"epoch": {"start_size": population_size}},
        "demes": [
            {"name": "ancestor0", "epochs": [{"end_time": 10}]},
            {"name": "ancestor1", "epochs": [{"end_time": 10}]},
        ]
        + [
            {
                "name": f"child_{j}",
                "ancestors": ["ancestor0", "ancestor1"],
                "proportions": [0.5, 0.5],
                "start_time": 10,
            }
            for j in range(num_demes)
        ],
    }
    return graph


class TestValidateGraph:
    def test_empty_document(self):
        with pytest.raises(KeyError):
            parser.parse({})

    def test_time_units(self):
        data = minimal_graph()
        data["time_units"] = 12
        with pytest.raises(TypeError):
            parser.parse(data)

    def test_generation_time_generations(self):
        data = minimal_graph()
        data["time_units"] = "generations"
        data["generation_time"] = 1
        graph = parser.parse(data)
        assert graph.generation_time == 1

    def test_bad_generation_time(self):
        data = minimal_graph()
        data["time_units"] = "generations"
        data["generation_time"] = 12
        with pytest.raises(ValueError):
            parser.parse(data)

    def test_generation_time(self):
        data = minimal_graph()
        data["time_units"] = "years"
        with pytest.raises(ValueError):
            parser.parse(data)
        data["generation_time"] = "string"
        with pytest.raises(TypeError):
            parser.parse(data)

    def test_doi(self):
        data = minimal_graph()
        data["doi"] = {"x:", "y"}
        with pytest.raises(TypeError):
            parser.parse(data)
        data["doi"] = [2134]
        with pytest.raises(TypeError):
            parser.parse(data)

    def test_description(self):
        data = minimal_graph()
        data["description"] = 1234
        with pytest.raises(TypeError):
            parser.parse(data)


class TestExtraFields:
    def validate(self, data):
        with pytest.raises(ValueError, match="Extra fields"):
            parser.parse(data)

    def test_top_level(self):
        data = minimal_graph()
        data["extra_field"] = 1234
        self.validate(data)

    def test_deme(self):
        data = minimal_graph()
        data["demes"][0]["extra_field"] = 1234
        self.validate(data)

    def test_epoch(self):
        data = single_deme_graph()
        data["demes"][0]["epochs"][0]["extra_field"] = 1234
        self.validate(data)

    def test_migration(self):
        data = island_model_graph(3)
        data["migrations"][0]["extra_field"] = 1234
        self.validate(data)

    def test_pulse(self):
        data = minimal_graph(2)
        data["pulses"] = [
            {
                "sources": ["deme0"],
                "dest": "deme1",
                "time": 1,
                "proportions": [0.5],
                "extra_field": 1234,
            }
        ]
        self.validate(data)

    def test_top_level_defaults(self):
        data = minimal_graph()
        data["defaults"] = {"extra_field": 1234}
        self.validate(data)

    def test_deme_defaults(self):
        data = minimal_graph()
        data["demes"][0]["defaults"] = {"extra_field": 1234}
        self.validate(data)


class TestExtraFieldsDefaults:
    def validate(self, data):
        with pytest.raises(ValueError, match="Only fields"):
            parser.parse(data)

    def test_deme(self):
        data = minimal_graph()
        data["defaults"] = {"deme": {"extra_field": 1234}}
        self.validate(data)

    def test_global_epoch(self):
        data = minimal_graph()
        data["defaults"] = {"epoch": {"extra_field": 1234}}
        self.validate(data)

    def test_migration(self):
        data = island_model_graph()
        data["defaults"] = {"migration": {"extra_field": 1234}}
        self.validate(data)
        # Should also be triggered when we have no migrations.
        data = minimal_graph()
        data["defaults"] = {"migration": {"extra_field": 1234}}
        self.validate(data)

    def test_pulse(self):
        data = minimal_graph()
        data["defaults"] = {"pulse": {"extra_field": 1234}}
        self.validate(data)


class TestValidateDeme:
    def test_id(self):
        data = minimal_graph()
        data["demes"][0]["name"] = 1234
        with pytest.raises(TypeError):
            parser.parse(data)
        data["demes"][0]["name"] = "not an identifier"
        with pytest.raises(ValueError):
            parser.parse(data)

    def test_description(self):
        data = minimal_graph()
        data["demes"][0]["description"] = 1234
        with pytest.raises(TypeError):
            parser.parse(data)

    def test_duplicate_deme_ids(self):
        data = minimal_graph(num_demes=2)
        data["demes"][0]["name"] = data["demes"][1]["name"]
        with pytest.raises(ValueError):
            parser.parse(data)

    def test_self_ancestor(self):
        data = minimal_graph(num_demes=2)
        data["demes"][0]["ancestors"] = ["deme1"]
        # Fails trying to find the deme with current ID.
        with pytest.raises(KeyError):
            parser.parse(data)

    def test_bad_proportions(self):
        data = single_ancestor_graph(1)
        for bad_proportion in [[], [0.9, 0.1]]:
            data["demes"][1]["proportions"] = bad_proportion
            with pytest.raises(ValueError, match="same length"):
                parser.parse(data)
        for bad_proportion in [[0.5]]:
            data["demes"][1]["proportions"] = bad_proportion
            with pytest.raises(ValueError, match="Sum of proportions"):
                parser.parse(data)

        data = two_ancestor_graph()
        for bad_proportion in [[0.6, 0.6]]:
            data["demes"][2]["proportions"] = bad_proportion
            with pytest.raises(ValueError, match="Sum of proportions"):
                parser.parse(data)

    def test_missing_time_two_ancestors(self):
        data = two_ancestor_graph()
        graph = parser.parse(data)
        assert len(graph.demes) == 3
        del data["demes"][2]["start_time"]
        with pytest.raises(ValueError, match="explicitly set Deme.start_time"):
            parser.parse(data)

    def test_missing_proportion_two_ancestors(self):
        data = two_ancestor_graph()
        graph = parser.parse(data)
        assert len(graph.demes) == 3
        del data["demes"][2]["proportions"]
        with pytest.raises(ValueError, match="Must specify proportions"):
            parser.parse(data)

    def test_time_travel(self):
        data = {
            "time_units": "generations",
            "defaults": {"epoch": {"start_size": 1}},
            "demes": [
                {"name": "deme0"},
                {
                    "name": "deme1",
                    "start_time": 100,
                    "ancestors": ["deme0"],
                    "epochs": [{"end_time": 20}],
                },
                {
                    "name": "deme2",
                    "start_time": 150,
                    "ancestors": ["deme1"],
                    "epochs": [{"end_time": 120}],
                },
            ],
        }
        with pytest.raises(ValueError):
            parser.parse(data)

    def test_root_with_finite_start_time(self):
        data = {
            "time_units": "generations",
            "demes": [
                {"name": "deme0", "start_time": 100, "epochs": [{"start_size": 1}]}
            ],
        }
        with pytest.raises(ValueError):
            parser.parse(data)

    def test_start_time_type(self):
        data = single_ancestor_graph()
        data["demes"][1]["start_time"] = "1000"
        with pytest.raises(TypeError):
            parser.parse(data)


class TestValidateEpoch:
    def test_end_time(self):
        data = minimal_graph()
        epoch = data["demes"][0]["epochs"][0]
        epoch["end_time"] = "x"
        with pytest.raises(TypeError):
            parser.parse(data)
        epoch["end_time"] = -1000
        with pytest.raises(ValueError):
            parser.parse(data)

    def test_no_end_time(self):
        data = single_deme_graph(num_epochs=3)
        graph = parser.parse(data)
        assert len(graph.demes["deme0"].epochs) == 3
        del data["demes"][0]["epochs"][1]["end_time"]
        with pytest.raises(ValueError, match="end_time must be specified"):
            parser.parse(data)

    def test_end_time_out_of_order(self):
        data = single_deme_graph(num_epochs=3)
        graph = parser.parse(data)
        assert len(graph.demes["deme0"].epochs) == 3
        data["demes"][0]["epochs"][1]["end_time"] = 0
        with pytest.raises(ValueError, match="end_times must be in decreasing"):
            parser.parse(data)

    def test_start_size(self):
        data = minimal_graph()
        epoch = data["demes"][0]["epochs"][0]
        epoch["start_size"] = "x"
        with pytest.raises(TypeError):
            parser.parse(data)
        epoch["start_size"] = -1000
        with pytest.raises(ValueError):
            parser.parse(data)

    def test_end_size(self):
        data = minimal_graph()
        epoch = data["demes"][0]["epochs"][0]
        epoch["end_size"] = "x"
        with pytest.raises(TypeError):
            parser.parse(data)
        epoch["end_size"] = -1000
        with pytest.raises(ValueError):
            parser.parse(data)

    def test_infinite_interval_non_constant(self):
        data = minimal_graph()
        epoch = data["demes"][0]["epochs"][0]
        epoch["end_size"] = 1000
        epoch["start_size"] = 1000.01
        with pytest.raises(ValueError):
            parser.parse(data)


class TestResolveEpochTimes:
    def test_minimal(self):
        graph = parser.parse(minimal_graph(population_size=10))
        deme = graph.demes["deme0"]
        assert deme.start_time == math.inf
        assert deme.end_time == 0
        assert len(deme.epochs) == 1
        epoch = deme.epochs[0]
        assert epoch.end_time == 0
        assert epoch.start_size == 10
        assert epoch.end_size == 10
        assert epoch.size_function == "constant"
        assert epoch.selfing_rate == 0
        assert epoch.cloning_rate == 0

    def test_line_topology_one_epoch(self):
        data = {
            "time_units": "generations",
            "defaults": {"epoch": {"start_size": 1}},
            "demes": [
                {"name": "deme0", "epochs": [{"end_time": 20}]},
                {"name": "deme1", "ancestors": ["deme0"], "epochs": [{"end_time": 10}]},
                {"name": "deme2", "ancestors": ["deme1"]},
            ],
        }
        graph = parser.parse(data)
        assert graph.demes["deme0"].start_time == math.inf
        assert graph.demes["deme0"].end_time == 20
        assert graph.demes["deme1"].start_time == 20
        assert graph.demes["deme1"].end_time == 10
        assert graph.demes["deme2"].start_time == 10
        assert graph.demes["deme2"].end_time == 0


class TestPulse:
    def test_simple(self):
        data = minimal_graph(num_demes=2)
        data["pulses"] = [
            {"sources": ["deme0"], "dest": "deme1", "proportions": [0.5], "time": 1}
        ]
        parsed = parser.parse(data).as_json_dict()
        assert data["pulses"] == parsed["pulses"]

    def test_multiple_sources(self):
        data = minimal_graph(num_demes=3)
        data["pulses"] = [
            {
                "sources": ["deme0", "deme1"],
                "dest": "deme2",
                "proportions": [0.1, 0.1],
                "time": 1,
            }
        ]
        parsed = parser.parse(data).as_json_dict()
        assert data["pulses"] == parsed["pulses"]

    def test_all_missing(self):
        data = minimal_graph(num_demes=2)
        data["pulses"] = [{}]
        with pytest.raises(KeyError):
            parser.parse(data)

    @pytest.mark.parametrize("proportion", [-1, 2])
    def test_bad_proportion(self, proportion):
        data = minimal_graph(num_demes=2)
        data["pulses"] = [
            {
                "sources": ["deme0"],
                "dest": "deme1",
                "proportions": [proportion],
                "time": 1,
            }
        ]
        with pytest.raises(ValueError):
            parser.parse(data)

    @pytest.mark.parametrize("proportions", [[], [0.1, 0.1]])
    def test_bad_proportions_length(self, proportions):
        data = minimal_graph(num_demes=2)
        data["pulses"] = [
            {
                "sources": ["deme0"],
                "dest": "deme1",
                "proportions": proportions,
                "time": 1,
            }
        ]
        with pytest.raises(
            ValueError, match="Sources and proportions must have same lengths"
        ):
            parser.parse(data)

    def test_bad_deme(self):
        data = minimal_graph(num_demes=2)
        data["pulses"] = [
            {"sources": ["deme3"], "dest": "deme1", "proportions": [0.5], "time": 1}
        ]
        with pytest.raises(KeyError):
            parser.parse(data)

    def test_empty_sources_list(self):
        data = minimal_graph(num_demes=2)
        data["pulses"] = [
            {"sources": [], "dest": "deme1", "proportions": [], "time": 1}
        ]
        with pytest.raises(ValueError, match="Must have one or more source demes"):
            parser.parse(data)

    def test_source_deme_equal_to_dest(self):
        data = minimal_graph(num_demes=2)
        data["pulses"] = [
            {"sources": ["deme0"], "dest": "deme0", "proportions": [0.5], "time": 1}
        ]
        with pytest.raises(ValueError, match="source deme equal to dest"):
            parser.parse(data)

    def test_duplicate_sources(self):
        data = minimal_graph(num_demes=2)
        data["pulses"] = [
            {
                "sources": ["deme0", "deme0"],
                "dest": "deme1",
                "proportions": [0.1, 0.1],
                "time": 1,
            }
        ]
        with pytest.raises(ValueError, match="Duplicate deme in sources"):
            parser.parse(data)

    def test_bad_time(self):
        data = minimal_graph(num_demes=2)
        data["demes"][1]["start_time"] = 10
        data["demes"][1]["ancestors"] = ["deme0"]
        data["pulses"] = [
            {"sources": ["deme0"], "dest": "deme1", "proportions": [0.5], "time": 20}
        ]
        with pytest.raises(ValueError, match="does not exist"):
            parser.parse(data)

    def test_pulse_at_deme_start_time(self):
        # The dest deme can receive a pulse at its start_time.
        data = minimal_graph(num_demes=3)
        data["demes"][1]["start_time"] = 10
        data["demes"][1]["ancestors"] = ["deme0"]
        data["pulses"] = [
            {"sources": ["deme2"], "dest": "deme1", "proportions": [0.5], "time": 10}
        ]
        parser.parse(data)

        # There can't be a pulse at the source deme's start_time.
        data["pulses"] = [
            {"sources": ["deme1"], "dest": "deme2", "proportions": [0.5], "time": 10}
        ]
        with pytest.raises(ValueError, match="does not exist"):
            parser.parse(data)

    def test_pulse_at_deme_end_time(self):
        # There can be a pulse at the source deme's end_time.
        data = minimal_graph(num_demes=2)
        data["demes"][1]["epochs"][0]["end_time"] = 10
        data["pulses"] = [
            {"sources": ["deme1"], "dest": "deme0", "proportions": [0.5], "time": 10}
        ]
        parser.parse(data)

        # The dest deme can't receive a pulse at its end_time.
        data["pulses"] = [
            {"sources": ["deme0"], "dest": "deme1", "proportions": [0.5], "time": 10}
        ]
        with pytest.raises(ValueError, match="does not exist"):
            parser.parse(data)

    def test_proportions_sum(self):
        data = minimal_graph(num_demes=3)
        data["pulses"] = [
            {"sources": ["deme1"], "dest": "deme0", "proportions": [0.6], "time": 10},
            {"sources": ["deme2"], "dest": "deme0", "proportions": [0.6], "time": 10},
        ]
        parser.parse(data)

    def test_bad_proportions_sum(self):
        data = minimal_graph(num_demes=3)
        data["pulses"] = [
            {
                "sources": ["deme1", "deme2"],
                "dest": "deme0",
                "proportions": [0.6, 0.6],
                "time": 10,
            },
        ]
        with pytest.raises(ValueError, match="sum to more than 1"):
            parser.parse(data)

    def test_pulse_order1(self):
        # Pulses given in time-descending order, so order is maintained.
        data = minimal_graph(num_demes=2)
        data["pulses"] = [
            {"sources": ["deme0"], "dest": "deme1", "proportions": [0.5], "time": 1.2},
            {"sources": ["deme1"], "dest": "deme0", "proportions": [0.5], "time": 1},
        ]
        parsed = parser.parse(data).as_json_dict()
        assert data["pulses"] == parsed["pulses"]

    def test_pulse_order2(self):
        # Pulses given in time-ascending order, so order should be reversed.
        data = minimal_graph(num_demes=2)
        data["pulses"] = [
            {"sources": ["deme0"], "dest": "deme1", "proportions": [0.5], "time": 1},
            {"sources": ["deme1"], "dest": "deme0", "proportions": [0.5], "time": 1.2},
        ]
        parsed = parser.parse(data).as_json_dict()
        assert data["pulses"][0] == parsed["pulses"][1]
        assert data["pulses"][1] == parsed["pulses"][0]

    def test_pulse_order3(self):
        # Pulses have the same time, so order is maintained.
        data = minimal_graph(num_demes=2)
        data["pulses"] = [
            {"sources": ["deme0"], "dest": "deme1", "proportions": [0.5], "time": 1},
            {"sources": ["deme1"], "dest": "deme0", "proportions": [0.5], "time": 1},
        ]
        parsed = parser.parse(data).as_json_dict()
        assert data["pulses"] == parsed["pulses"]

        # Switch the given ordering, to double check that we didn't maintain
        # order by accident (e.g. by sorting on source name or something crazy).
        data["pulses"] = [
            {"sources": ["deme1"], "dest": "deme0", "proportions": [0.5], "time": 1},
            {"sources": ["deme0"], "dest": "deme1", "proportions": [0.5], "time": 1},
        ]
        parsed = parser.parse(data).as_json_dict()
        assert data["pulses"] == parsed["pulses"]


class TestMigration:
    def test_simple_asymmetric(self):
        data = minimal_graph(num_demes=2)
        data["migrations"] = [
            {
                "source": "deme0",
                "dest": "deme1",
                "rate": 0.5,
                "start_time": 2,
                "end_time": 1,
            }
        ]
        parsed = parser.parse(data).as_json_dict()
        assert data["migrations"] == parsed["migrations"]

    def test_simple_asymmetric_default_times(self):
        data = minimal_graph(num_demes=2)
        data["migrations"] = [{"source": "deme0", "dest": "deme1", "rate": 0}]
        parsed = parser.parse(data).as_json_dict()
        assert parsed["migrations"] == [
            {
                "source": "deme0",
                "dest": "deme1",
                "rate": 0,
                "start_time": "Infinity",
                "end_time": 0,
            }
        ]

    def test_simple_symmetric(self):
        data = minimal_graph(num_demes=2)
        data["migrations"] = [
            {"demes": ["deme0", "deme1"], "rate": 0.5, "start_time": 2, "end_time": 1}
        ]
        parsed = parser.parse(data).as_json_dict()
        assert len(parsed["migrations"]) == 2
        assert {
            "source": "deme0",
            "dest": "deme1",
            "rate": 0.5,
            "start_time": 2,
            "end_time": 1,
        } in parsed["migrations"]
        assert {
            "source": "deme1",
            "dest": "deme0",
            "rate": 0.5,
            "start_time": 2,
            "end_time": 1,
        } in parsed["migrations"]

    def test_pairwise_resolution_for_symmetric(self):
        data = minimal_graph(num_demes=3)
        data["demes"][1]["epochs"] = [{"start_size": 1, "end_time": 50}]
        data["demes"][2]["ancestors"] = ["deme1"]
        data["demes"][2]["start_time"] = 100
        data["migrations"] = [{"demes": ["deme0", "deme1", "deme2"], "rate": 0.5}]
        parsed = parser.parse(data).as_json_dict()
        assert len(parsed["migrations"]) == 6
        assert {
            "source": "deme0",
            "dest": "deme1",
            "rate": 0.5,
            "start_time": "Infinity",
            "end_time": 50,
        } in parsed["migrations"]
        assert {
            "source": "deme1",
            "dest": "deme0",
            "rate": 0.5,
            "start_time": "Infinity",
            "end_time": 50,
        } in parsed["migrations"]
        assert {
            "source": "deme0",
            "dest": "deme2",
            "rate": 0.5,
            "start_time": 100,
            "end_time": 0,
        } in parsed["migrations"]
        assert {
            "source": "deme2",
            "dest": "deme0",
            "rate": 0.5,
            "start_time": 100,
            "end_time": 0,
        } in parsed["migrations"]
        assert {
            "source": "deme1",
            "dest": "deme2",
            "rate": 0.5,
            "start_time": 100,
            "end_time": 50,
        } in parsed["migrations"]
        assert {
            "source": "deme2",
            "dest": "deme1",
            "rate": 0.5,
            "start_time": 100,
            "end_time": 50,
        } in parsed["migrations"]

    def test_symmetric_and_asymmetric(self):
        data = minimal_graph(num_demes=2)
        data["migrations"] = [
            {
                "source": "deme0",
                "dest": "deme1",
                "demes": ["deme0", "deme1"],
                "rate": 0.5,
            }
        ]
        with pytest.raises(ValueError, match="either source and dest, or demes"):
            parser.parse(data)

        data = minimal_graph(num_demes=2)
        data["migrations"] = [
            {
                "dest": "deme1",
                "demes": ["deme0", "deme1"],
                "rate": 0.5,
            }
        ]
        with pytest.raises(ValueError, match="either source and dest, or demes"):
            parser.parse(data)

        data = minimal_graph(num_demes=2)
        data["migrations"] = [
            {
                "source": "deme0",
                "demes": ["deme0", "deme1"],
                "rate": 0.5,
            }
        ]
        with pytest.raises(ValueError, match="either source and dest, or demes"):
            parser.parse(data)

    def test_neither_symmetric_or_asymmetric(self):
        data = minimal_graph(num_demes=2)
        data["migrations"] = [
            {
                "rate": 0.5,
            }
        ]
        with pytest.raises(ValueError, match="either source and dest, or demes"):
            parser.parse(data)

    def test_same_deme_asymmetric(self):
        data = minimal_graph(num_demes=2)
        data["migrations"] = [
            {
                "source": "deme0",
                "dest": "deme0",
                "rate": 0.5,
            }
        ]
        with pytest.raises(ValueError, match="migrate from a deme to itself"):
            parser.parse(data)

    def test_same_deme_symmetric(self):
        data = minimal_graph(num_demes=2)
        data["migrations"] = [
            {
                "demes": ["deme0", "deme0"],
                "rate": 0.5,
            }
        ]
        with pytest.raises(ValueError, match="migrate from a deme to itself"):
            parser.parse(data)

    def test_bad_start_time_value(self):
        data = minimal_graph(num_demes=2)
        data["migrations"] = [
            {"demes": ["deme0", "deme1"], "rate": 0.5, "start_time": -1, "end_time": 1}
        ]
        with pytest.raises(ValueError):
            parser.parse(data).as_json_dict()

    def test_bad_end_time_value(self):
        data = minimal_graph(num_demes=2)
        data["migrations"] = [
            {"demes": ["deme0", "deme1"], "rate": 0.5, "start_time": 2, "end_time": -1}
        ]
        with pytest.raises(ValueError):
            parser.parse(data).as_json_dict()

    def test_bad_end_time_interval(self):
        data = minimal_graph(num_demes=2)
        data["migrations"] = [
            {"demes": ["deme0", "deme1"], "rate": 0.5, "start_time": 1, "end_time": 100}
        ]
        with pytest.raises(ValueError, match="start_time must be > end_time"):
            parser.parse(data).as_json_dict()

    def test_bad_time_asymmetric(self):
        data = minimal_graph(num_demes=2)
        data["demes"][1]["start_time"] = 10
        data["demes"][1]["ancestors"] = ["deme0"]
        data["migrations"] = [
            {
                "source": "deme0",
                "dest": "deme1",
                "rate": 0.5,
                "start_time": 20,
                "end_time": 11,
            }
        ]
        with pytest.raises(ValueError, match="time interval"):
            parser.parse(data)

    def test_bad_time_symmetric(self):
        data = minimal_graph(num_demes=2)
        data["demes"][1]["start_time"] = 10
        data["demes"][1]["ancestors"] = ["deme0"]
        data["migrations"] = [
            {"demes": ["deme0", "deme1"], "rate": 0.5, "start_time": 20, "end_time": 11}
        ]
        with pytest.raises(ValueError, match="time interval"):
            parser.parse(data)

    def test_bad_migration_competing_migrations(self):
        data = minimal_graph(num_demes=2)
        data["migrations"] = [
            {
                "demes": ["deme0", "deme1"],
                "rate": 0.5,
                "start_time": 20,
                "end_time": 11,
            },
            {"demes": ["deme0", "deme1"], "rate": 0.5, "start_time": 12, "end_time": 1},
        ]
        with pytest.raises(ValueError, match="Competing migration definitions"):
            parser.parse(data)

    def test_bad_migration_rates_sum_to_more_than_1(self):
        data = minimal_graph(num_demes=3)
        data["migrations"] = [
            {
                "source": "deme0",
                "dest": "deme2",
                "rate": 0.6,
                "start_time": 20,
                "end_time": 11,
            },
            {
                "source": "deme1",
                "dest": "deme2",
                "rate": 0.6,
                "start_time": 12,
                "end_time": 1,
            },
        ]
        with pytest.raises(ValueError, match="sum to more than 1"):
            parser.parse(data)


class TestResolveEpochSizes:
    def test_single_epoch(self):
        data = minimal_graph()
        data["demes"][0]["epochs"][0] = {}
        with pytest.raises(ValueError):
            parser.parse(data)
        data["demes"][0]["epochs"][0] = {"start_size": 100}
        resolved = parser.parse(data).as_json_dict()
        assert resolved["demes"][0]["epochs"][0]["start_size"] == 100
        assert resolved["demes"][0]["epochs"][0]["end_size"] == 100

        data["demes"][0]["epochs"][0] = {"end_size": 200}
        resolved = parser.parse(data).as_json_dict()
        assert resolved["demes"][0]["epochs"][0]["start_size"] == 200
        assert resolved["demes"][0]["epochs"][0]["end_size"] == 200

    def test_propagate_start_size(self):
        data = minimal_graph()
        data["demes"][0]["epochs"] = [
            {"start_size": 1, "end_time": 100},
            {"end_size": 2, "end_time": 50},
        ]

        resolved = parser.parse(data).as_json_dict()
        assert resolved["demes"][0]["epochs"][0]["start_size"] == 1
        assert resolved["demes"][0]["epochs"][0]["end_size"] == 1
        assert resolved["demes"][0]["epochs"][1]["start_size"] == 1
        assert resolved["demes"][0]["epochs"][1]["end_size"] == 2

    def test_propagate_start_and_end_size(self):
        data = minimal_graph()
        data["demes"][0]["epochs"] = [
            {"start_size": 1, "end_time": 100},
            {"end_time": 50},
            {},
        ]
        resolved = parser.parse(data).as_json_dict()
        for j in range(3):
            assert resolved["demes"][0]["epochs"][j]["start_size"] == 1
            assert resolved["demes"][0]["epochs"][j]["end_size"] == 1


class TestDefaults:
    def test_asymmetric_migration_changing_rate(self):
        data = minimal_graph(num_demes=2)
        data["defaults"] = {"migration": {"source": "deme0", "dest": "deme1"}}
        data["migrations"] = [
            {"start_time": 2, "end_time": 1, "rate": 1},
            {"start_time": 1, "end_time": 0, "rate": 0.5},
        ]
        graph = parser.parse(data)
        assert len(graph.migrations) == 2
        for migration in graph.migrations:
            assert migration.source.name == "deme0"
            assert migration.dest.name == "deme1"
        assert graph.migrations[0].rate == 1
        assert graph.migrations[0].start_time == 2
        assert graph.migrations[0].end_time == 1
        assert graph.migrations[1].rate == 0.5
        assert graph.migrations[1].start_time == 1
        assert graph.migrations[1].end_time == 0

    def test_symmetric_migration_changing_rate(self):
        data = minimal_graph(num_demes=2)
        data["defaults"] = {"migration": {"demes": ["deme0", "deme1"]}}
        data["migrations"] = [
            {"start_time": 2, "end_time": 1, "rate": 1},
            {"start_time": 1, "end_time": 0, "rate": 0.5},
        ]
        graph = parser.parse(data)
        parsed = graph.as_json_dict()
        assert len(parsed["migrations"]) == 4
        assert {
            "source": "deme0",
            "dest": "deme1",
            "start_time": 2,
            "end_time": 1,
            "rate": 1,
        } in parsed["migrations"]
        assert {
            "source": "deme1",
            "dest": "deme0",
            "start_time": 2,
            "end_time": 1,
            "rate": 1,
        } in parsed["migrations"]
        assert {
            "source": "deme0",
            "dest": "deme1",
            "start_time": 1,
            "end_time": 0,
            "rate": 0.5,
        } in parsed["migrations"]
        assert {
            "source": "deme1",
            "dest": "deme0",
            "start_time": 1,
            "end_time": 0,
            "rate": 0.5,
        } in parsed["migrations"]

    def test_migration_start_time_end_time_rate(self):
        data = minimal_graph(num_demes=3)
        data["defaults"] = {"migration": {"start_time": 2, "end_time": 1, "rate": 1}}
        data["migrations"] = [
            {"source": "deme0", "dest": "deme1"},
            {"source": "deme1", "dest": "deme2"},
        ]
        graph = parser.parse(data)
        assert len(graph.migrations) == 2
        for migration in graph.migrations:
            migration.start_time == 2
            migration.end_time == 1
            migration.rate == 1
        assert graph.migrations[0].source.name == "deme0"
        assert graph.migrations[0].dest.name == "deme1"
        assert graph.migrations[1].source.name == "deme1"
        assert graph.migrations[1].dest.name == "deme2"

    def test_pulse_time_proportion(self):
        data = minimal_graph(num_demes=3)
        data["defaults"] = {"pulse": {"time": 1, "proportions": [0.5]}}
        data["pulses"] = [
            {"sources": ["deme0"], "dest": "deme1"},
            {"sources": ["deme1"], "dest": "deme2"},
        ]
        graph = parser.parse(data)
        assert len(graph.pulses) == 2
        for pulse in graph.pulses:
            pulse.time == 1
            pulse.proportions == [0.5]
        assert len(graph.pulses[0].sources) == 1
        assert graph.pulses[0].sources[0].name == "deme0"
        assert graph.pulses[0].dest.name == "deme1"
        assert len(graph.pulses[1].sources) == 1
        assert graph.pulses[1].sources[0].name == "deme1"
        assert graph.pulses[1].dest.name == "deme2"

    def test_deme_description_start_time(self):
        data = minimal_graph(num_demes=3)
        data["defaults"] = {
            "deme": {"description": "default", "start_time": 1, "ancestors": ["deme0"]}
        }
        data["demes"][0]["start_time"] = math.inf
        data["demes"][0]["ancestors"] = []
        data["demes"][1]["description"] = "not default"
        data["demes"][1]["start_time"] = 2
        parsed = parser.parse(data).as_json_dict()
        assert parsed["demes"][1]["description"] == "not default"
        assert parsed["demes"][1]["start_time"] == 2
        assert parsed["demes"][2]["description"] == "default"
        assert parsed["demes"][2]["start_time"] == 1

    def test_deme_ancestors_proportions(self):
        data = minimal_graph(num_demes=4)
        data["defaults"] = {
            "deme": {
                "ancestors": ["deme0", "deme1"],
                "proportions": [0.5, 0.5],
                "start_time": 1,
            }
        }
        for j in range(2):
            data["demes"][j]["start_time"] = math.inf
            data["demes"][j]["ancestors"] = []
            data["demes"][j]["proportions"] = []
        parsed = parser.parse(data).as_json_dict()
        assert len(parsed["demes"]) == 4
        assert parsed["demes"][0]["ancestors"] == []
        assert parsed["demes"][1]["ancestors"] == []
        assert parsed["demes"][2]["ancestors"] == ["deme0", "deme1"]
        assert parsed["demes"][3]["ancestors"] == ["deme0", "deme1"]
        assert parsed["demes"][2]["proportions"] == [0.5, 0.5]
        assert parsed["demes"][3]["proportions"] == [0.5, 0.5]

    def test_no_epochs_specified(self):
        num_demes = 5
        data = {
            "time_units": "generations",
            "defaults": {"epoch": {"start_size": 1, "end_size": 2, "end_time": 10}},
            "demes": [{"name": "deme0", "epochs": [{"start_size": 1, "end_size": 1}]}]
            + [
                {"name": f"deme{j}", "start_time": 100, "ancestors": ["deme0"]}
                for j in range(1, num_demes)
            ],
        }
        graph = parser.parse(data)
        assert len(graph.demes) == num_demes
        for deme in list(graph.demes.values())[1:]:
            assert len(deme.epochs) == 1
            epoch = deme.epochs[0]
            assert epoch.start_size == 1
            assert epoch.end_size == 2
            assert epoch.end_time == 10

    def test_many_epochs_one_deme_global(self):
        num_epochs = 4
        data = {
            "time_units": "generations",
            "defaults": {
                "epoch": {
                    "start_size": 1,
                    "end_size": 2,
                    "cloning_rate": 0.5,
                    "selfing_rate": 0.1,
                }
            },
            "demes": [
                {
                    "name": "deme0",
                    "epochs": [{"end_time": 100, "start_size": 1, "end_size": 1}]
                    + [{"end_time": j} for j in range(num_epochs - 1, -1, -1)],
                }
            ],
        }
        graph = parser.parse(data)
        deme = graph.demes["deme0"]
        assert len(deme.epochs) == num_epochs + 1
        for epoch in deme.epochs[1:]:
            assert epoch.start_size == 1
            assert epoch.end_size == 2
            assert epoch.cloning_rate == 0.5
            assert epoch.selfing_rate == 0.1

    def test_many_epochs_one_deme_local(self):
        num_epochs = 4
        data = {
            "time_units": "generations",
            "demes": [
                {
                    "name": "deme0",
                    "defaults": {
                        "epoch": {
                            "start_size": 1,
                            "end_size": 2,
                            "cloning_rate": 0.5,
                            "selfing_rate": 0.1,
                        }
                    },
                    "epochs": [{"end_time": 100, "start_size": 1, "end_size": 1}]
                    + [{"end_time": j} for j in range(num_epochs - 1, -1, -1)],
                }
            ],
        }
        graph = parser.parse(data)
        deme = graph.demes["deme0"]
        assert len(deme.epochs) == num_epochs + 1
        for epoch in deme.epochs[1:]:
            assert epoch.start_size == 1
            assert epoch.end_size == 2
            assert epoch.cloning_rate == 0.5
            assert epoch.selfing_rate == 0.1

    def test_epoch_default_overrides(self):
        num_epochs = 4
        data = {
            "time_units": "generations",
            "defaults": {
                "deme": {"start_time": 100, "ancestors": ["ancestral"]},
                "epoch": {
                    "start_size": 1,
                    "end_size": 2,
                    "cloning_rate": 0.5,
                    "selfing_rate": 0.1,
                },
            },
            "demes": [
                {
                    "name": "ancestral",
                    "start_time": math.inf,
                    "ancestors": [],
                    "epochs": [{"start_size": 1, "end_size": 1}],
                },
                {
                    "name": "deme0",
                    "defaults": {
                        "epoch": {
                            "start_size": 2,
                            "end_size": 3,
                            "cloning_rate": 0.6,
                            "selfing_rate": 0.2,
                        }
                    },
                    "epochs": [{"end_time": j} for j in range(num_epochs - 1, -1, -1)],
                },
                {
                    "name": "deme1",
                    "epochs": [{"end_time": j} for j in range(num_epochs - 1, -1, -1)],
                },
                {
                    "name": "deme2",
                    "epochs": [{"end_time": j} for j in range(num_epochs - 1, -1, -1)],
                },
            ],
        }
        graph = parser.parse(data)
        deme = graph.demes["deme0"]
        assert len(deme.epochs) == num_epochs
        for epoch in deme.epochs:
            assert epoch.start_size == 2
            assert epoch.end_size == 3
            assert epoch.cloning_rate == 0.6
            assert epoch.selfing_rate == 0.2
        for deme_id in ["deme1", "deme2"]:
            deme = graph.demes[deme_id]
            assert len(deme.epochs) == num_epochs
            for epoch in deme.epochs:
                assert epoch.start_size == 1
                assert epoch.end_size == 2
                assert epoch.cloning_rate == 0.5
                assert epoch.selfing_rate == 0.1


class TestMetadata:
    def test_toplevel_metadata(self):
        data = minimal_graph()
        metadata = dict(foo=1, bar="two", nested=dict(things=dict(baz="baz")))
        data["metadata"] = metadata
        graph = parser.parse(data)
        assert graph.metadata == metadata

    @pytest.mark.parametrize("metadata", [None, 1, "string", [1, 2, 3]])
    def test_bad_toplevel_metadata(self, metadata):
        data = minimal_graph()
        data["metadata"] = metadata
        with pytest.raises(TypeError):
            parser.parse(data)


class TestGraphUtilities:
    def test_str(self):
        graph = parser.parse(minimal_graph())
        assert len(str(graph)) > 0


@pytest.mark.parametrize(
    "yaml_path", map(str, pathlib.Path("../examples/").glob("*.yaml"))
)
def test_examples(yaml_path):
    yaml = YAML(typ="safe")
    with open(yaml_path, encoding="utf-8") as source:
        data = yaml.load(source)
    graph = parser.parse(data)
    graph_data = graph.as_json_dict()

    yaml_path = pathlib.Path(yaml_path)
    json_path = yaml_path.parent / yaml_path.with_suffix(".resolved.json")
    with open(json_path, encoding="utf-8") as source:
        json_data = json.load(source)
    # Note: we'll probably need to do something less strict here.
    assert json_data == graph_data

    graph_copy = parser.parse(json_data)
    assert graph_copy == graph


class TestValidCases:
    def parse_file(self, yaml_path):
        yaml = YAML(typ="safe")
        with open(yaml_path, encoding="utf-8") as source:
            data = yaml.load(source)
        return parser.parse(data).as_json_dict()

    @pytest.mark.parametrize(
        "yaml_path", map(str, pathlib.Path("../test-cases/valid").glob("*.yaml"))
    )
    def test_resolve_equal(self, yaml_path):
        resolved = self.parse_file(yaml_path)
        # Reparsing the fully resolved model should give identical output.
        assert resolved == parser.parse(resolved).as_json_dict()

    @pytest.mark.parametrize(
        "yaml_path", map(str, pathlib.Path("../test-cases/valid").glob("*.yaml"))
    )
    def test_validates_base_schema(self, yaml_path):
        resolved = self.parse_file(yaml_path)
        yaml = YAML(typ="safe")
        with open("../schema/hdm-v1.0.yaml", encoding="utf-8") as source:
            schema = yaml.load(source)
        jsonschema.validate(instance=resolved, schema=schema)

    @pytest.mark.parametrize(
        "yaml_path", map(str, pathlib.Path("../test-cases/valid").glob("*.yaml"))
    )
    def test_validates_fully_qualified_schema(self, yaml_path):
        resolved = self.parse_file(yaml_path)
        yaml = YAML(typ="safe")
        with open("../schema/mdm-v1.0.yaml", encoding="utf-8") as source:
            schema = yaml.load(source)
        jsonschema.validate(instance=resolved, schema=schema)


@pytest.mark.parametrize(
    "yaml_path", map(str, pathlib.Path("../test-cases/invalid").glob("*.yaml"))
)
def test_invalid_testcases(yaml_path):
    yaml = YAML(typ="safe")
    with open(yaml_path, encoding="utf-8") as source:
        if yaml_path.endswith("invalid_fields_11.yaml"):
            # Weird case that's caught in ruamel.
            with pytest.raises(ConstructorError):
                data = yaml.load(source)
        else:
            data = yaml.load(source)
            with pytest.raises((ValueError, TypeError, KeyError)):
                parser.parse(data)



import glob
import pickle
import warnings

import numpy as np
import pytest

import demes
import fwdpy11


def test_invalid_demes_spec_modules():
    for bad in glob.glob("tests/demes-spec/test-cases/invalid/*.yaml"):
        with open(bad, "r") as f:
            yaml = "".join(f.readlines())
            # NOTE: cannot test these YAML with demes.Graph b/c
            # they cannot create a valid Graph.
            with pytest.raises(Exception):  # NOQA
                _ = fwdpy11.ForwardDemesGraph.from_demes(yaml, burnin=0)


def test_valid_demes_spec_modules():
    # Some valid models contain ambiguous pulses,
    # so we suppress those warnings here
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        for good in glob.glob("tests/demes-spec/test-cases/valid/*.yaml"):
            graph = demes.load(good)
            non_integer_deme_sizes = []
            for deme in graph.demes:
                for epoch in deme.epochs:
                    for size in [epoch.start_size, epoch.end_size]:
                        if np.isfinite(size) and np.modf(size)[0] != 0.0:
                            non_integer_deme_sizes.append(size)
            if len(non_integer_deme_sizes) == 0:
                with open(good, "r") as f:
                    yaml = "".join(f.readlines())
                    _ = fwdpy11.ForwardDemesGraph.from_demes(yaml, burnin=0)
                    _ = fwdpy11.ForwardDemesGraph.from_demes(
                        demes.loads(yaml), burnin=0)
                    with pytest.raises(TypeError):
                        _ = fwdpy11.ForwardDemesGraph.from_demes(
                            yaml, burnin=-1)
            else:
                with open(good, "r") as f:
                    yaml = "".join(f.readlines())
                    with pytest.raises(ValueError):
                        _ = fwdpy11.ForwardDemesGraph.from_demes(
                            yaml, burnin=0, round_non_integer_sizes=False)
                    if all(i > 0.5 for i in non_integer_deme_sizes):
                        for round_it in [True]:
                            _ = fwdpy11.ForwardDemesGraph.from_demes(
                                yaml, burnin=0, round_non_integer_sizes=round_it)
                            _ = fwdpy11.discrete_demography.from_demes(
                                good, burnin=0, round_non_integer_sizes=round_it)
                        for round_it in [None, False]:
                            with pytest.raises(ValueError):
                                _ = fwdpy11.ForwardDemesGraph.from_demes(
                                    yaml, burnin=0, round_non_integer_sizes=round_it)
                                _ = fwdpy11.discrete_demography.from_demes(
                                    good, burnin=0, round_non_integer_sizes=round_it)
                    else:
                        for round_it in [None, False]:
                            with pytest.raises(ValueError):
                                _ = fwdpy11.ForwardDemesGraph.from_demes(
                                    yaml, burnin=0, round_non_integer_sizes=round_it)
                                _ = fwdpy11.discrete_demography.from_demes(
                                    good, burnin=0, round_non_integer_sizes=round_it)


def test_pickle():
    yaml = """
time_units: generations
demes:
- name: deme1
  start_time: .inf
  epochs:
  - {end_size: 1200.0, end_time: 0, start_size: 1200.0}
- name: deme2
  start_time: .inf
  epochs:
  - {end_size: 100, end_time: 400.0, start_size: 100}
pulses:
- {sources: [deme2], dest: deme1, proportions: [0.09999999999999998], time: 400.0}
migrations: []
    """
    g = fwdpy11.ForwardDemesGraph.from_demes(
        yaml, 93, burnin_is_exact=True, round_non_integer_sizes=True)
    p = pickle.dumps(g)
    up = pickle.loads(p)
    assert g == up
    assert g.graph == up.graph

    # NOTE: we are testing an internal detail here:
    # pickling works via asdict and we don't
    # want to include the demes.Graph in that object.
    d = g.asdict()
    assert "graph" not in d
    assert "yaml" in d



import os
import sysconfig

import ycm_core

flags = [
    "-x",
    "c++",
    f"-I{os.path.dirname(os.path.abspath(__file__))}/fwdpy11/headers",
    f"-I{os.path.dirname(os.path.abspath(__file__))}/fwdpy11/headers/fwdpp",
    f"-I{sysconfig.get_paths()['include']}",
    "-I/usr/local/include",
    "-Wall",
    "-Werr",
    "-Wconversion",
    "-Weffc++",
    "-std=c++14",
]


def Settings(**kwargs):
    if kwargs["language"] == "cfamily":
        return {"flags": flags}
    if kwargs["language"] == "python":
        pass



import pickle
import unittest

import fwdpy11
from ezparams import mslike

import custom_additive as ca
import custom_stateless_genotype as general


class testCustomAdditive(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.pop = fwdpy11.DiploidPopulation(1000, 1.0)
        self.pdict = mslike(
            self.pop, dfe=fwdpy11.ExpS(0, 1, 1, -0.05),
            pneutral=0.95, simlen=10
        )
        self.pdict["gvalue"] = ca.additive()
        self.pdict["demography"] = \
            fwdpy11.ForwardDemesGraph.tubes(self.pop.deme_sizes()[1],
                                            burnin=self.pdict["simlen"],
                                            burnin_is_exact=True)
        self.rng = fwdpy11.GSLrng(42)
        self.params = fwdpy11.ModelParams(**self.pdict)

    def testEvolve(self):
        fwdpy11.evolvets(self.rng, self.pop, self.params, 10)

    def testPickle(self):
        a = self.params.gvalue
        p = pickle.dumps(a, -1)
        up = pickle.loads(p)
        self.assertEqual(type(a), type(up))


class testGeneralModule(unittest.TestCase):
    @classmethod
    def setUp(self):
        self.pop = fwdpy11.DiploidPopulation(1000, 1.0)
        self.pdict = mslike(
            self.pop,
            dfe=fwdpy11.ConstantS(0, 1, 1, -0.05, 0.05),
            pneutral=0.95,
            simlen=10,
        )
        self.pdict["gvalue"] = general.GeneralW()
        self.pdict["demography"] =\
            fwdpy11.ForwardDemesGraph.tubes(self.pop.deme_sizes()[1],
                                            burnin=self.pdict["simlen"],
                                            burnin_is_exact=True)
        self.rng = fwdpy11.GSLrng(42)
        self.params = fwdpy11.ModelParams(**self.pdict)

    def testPickle(self):
        a = self.params.gvalue
        p = pickle.dumps(a, -1)
        up = pickle.loads(p)
        self.assertEqual(type(a), type(up))

    def testEvolve(self):
        fwdpy11.evolvets(self.rng, self.pop, self.params, 10)


if __name__ == "__main__":
    unittest.main()
